/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

#pragma ident	"@(#)args.c	1.53	09/05/26 SMI"

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Based on a general purpose argument and configuration handling routines.
 * This is macro hell. This all works by including the file arg_vals.h
 * multiple times each time having defined the macros:
 *
 * SET_OPENING_BLURB(A)  A is a string that will get printed at the
 *	top of any output file containing the options.
 * START_SECTION(A, B)	Used to group options. A is the name of the
 *	group and B is a string that describes the section.
 * END_SECTION(A) Used to end a group of options. The A should match
 *	the A of START_SECTION above.
 *
 *
 * STARG(X, B, C)   to handle strings. The string "NULL" is special.
 *	Do not use it as one of your options.
 * ILIST(A, B, C) to handle comma lists with bracketed weightings.
 * BARG(X, B, C)	to handle single bits, boolean types.
 * CARG(X, B, C)	to handle chars.
 * SARG(X, B, C)	to handle shorts.
 * IARG(X, B, C)	to handle ints.
 * LLARG(X, B, C)	to handle long longs.
 * LARG(X, B, C)	to handle longs.
 *
 * the arguments to the args macros are all the same:
 *
 *	X	the name of the option.  This must be unique and will be the
 *		name of the element in the structure.
 *	B	the default value. This must be of the correct type for the
 *		macro.
 *	C	A string that describes the argument. This forms the coment
 *		used in the options files and when the usage messages is
 * 		printed.  It should not contain any special chars. Eg \n or
 *		\t or \r.
 *
 * After each include the macros get undefined. This way to add a configuration
 * option the only file you edit it arg_vals.h. What is more this args.c and
 * args.h can and are used by many different programs to generate many
 * different option structures.
 *
 * All your program has to do is call do_args(argc, argv). The program
 * then will take the standard four options:
 *
 *	-f option_file
 *		Causes options to be read from the named file.
 *	-o option=foo
 *		causes a specific option to be set.
 *	-h file
 *		Print the HTML usage into the file named, using the file '-'
 *		caused the HTML to be written to standard out and exit.
 *	-d file
 *		Print the current values of the options into the named file.
 *	-?
 *		Print all the options to standard out.
 *
 *	These options can be combined and are actioned in sequence.
 *
 *	prog -f foo -o bar=plap -d one -o bar=spam -d two
 *
 *	Reads all the options from the file foo, then sets the option
 *	bar to be plap, then prints the settings into the file "one"
 *	then sets bar to be spame and prints the options into the file
 *	"two".
 *
 * Additionally all options are stored in an option store so that
 * shared libraries can have options that are not known about by
 * diskomizer at compile time. The store is simply a sorted list.
 *
 * For reasonable numbers of options the search speed is acceptable
 * it is assumed that the consumers of options will cache their values
 * to minimize the number of calls into the option_store.
 *
 * See the comment in args.h for usage of the option store from shared
 * libraries.
 */
#include "arg_inc.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <diskomizer/log.h>
#include <usage_tracking/usage_tracking.h>
#include "disko_usage_track.h"
#include "time.h"

#define	COMMA ','
#define	DOLLAR '$'
#define	OPEN_BRACE '{'
#define	CLOSE_BRACE '}'
#define	OPEN_BRACKET '('
#define	CLOSE_BRACKET ')'
#define	LINE_LEN 75

#define	DISKOMIZER_PATH "DISKOMIZER_PATH"

static const char OPEN_BRACKET_STR[] = "(,";
static const char CLOSE_BRACKET_STR[] = ")";

#include "args.h"

static const char def_diskomizer_path[] = ":/opt/SUNWdiskomizer/etc";
static const char nil_str[] = "(nil)";

#define	NOT_NULL_FREE(X) if (X != NULL) free(X);
#define	NIL_STR(X) (X != NULL ? X : nil_str)
#include <libintl.h>

typedef long long (*operator_t)(long long x, long long y);
static long long
	(*choose_operator(char *str, char **newstr))(long long x, long long y);
static long long set(long long x, long long y);
/*
 * These structures are used to keep track of state information for the
 * parsing  of lines so that correct line numbers can be reported when
 * errors occur.
 */
struct lineinfo {
	int start;
	int end;
};
struct option_info {
	char *filename;
	char *line;
	uint_t no_report:1;
	uint_t had_error:1;
	uint_t built_in_option:1;
	uint_t used:1;
	const char *endchrs;
	struct lineinfo lineno;
};

struct option_entry {
	const char *key;
	int len;
	char *value;
	long long lv;
	struct option_info info;
};

struct option_store {
	int refcount;
	int nentries;
	struct option_entry *entries;
};
struct option_store opt_store;


static void print_html(const char *prog, const char *filename);
static void print_defaults(FILE *f);
static int print_option_file(const char *prog, const char *file, FILE *f);
static void set_defaults(void);
static int process_option(char *option, struct option_info *infop);
static int print_file(const char *prog, const char *filename,
	int (*print_func)(const char *, const char *, FILE *));

static option_status_t get_long_option(const char *key, long *out);
static option_status_t get_long_long_option(const char *key, long long *out);
static long long do_maths(long long cur, char *str, char **nstr,
		struct option_info *infop);
static void print_all_option_entries(FILE *f);

#define	STARG(A, B, C)
#define	ILIST(A, B, C)
#define	SARG(A, B, C)
#define	LLARG(A, B, C)
#define	LARG(A, B, C)
#define	CARG(A, B, C)
#define	BARG(A, B, C)
#define	IARG(A, B, C)
#define	START_SECTION(A, B)
#define	END_SECTION(A)
#define	SET_OPENING_BLURB(A) const char OPENING_STR[] = A;
#include "arg_vals.h"
#undef START_SECTION
#undef SET_OPENING_BLURB
#define	SET_OPENING_BLURB(A)
struct opt_sections {
#define	START_SECTION(A, B) char *A;
#include "arg_vals.h"
#undef START_SECTION
};
static struct opt_sections opts_scomments = {
#define	START_SECTION(A, B) B,
#include "arg_vals.h"
#undef START_SECTION
};
static struct opt_sections opts_snames = {
#define	START_SECTION(A, B) #A,
#include "arg_vals.h"

};
#undef STARG
#undef ILIST
#undef SARG
#undef LLARG
#undef LARG
#undef CARG
#undef BARG
#undef IARG
#undef START_SECTION
#undef END_SECTION
#undef SET_OPENING_BLURB

#define	START_SECTION(A, B)
#define	END_SECTION(A)
#define	SET_OPENING_BLURB(A)
struct options opts;
struct opts_comments {
#define	SARG(B, D, C) char *B;
#define	STARG SARG
#define	IARG SARG
#define	CARG SARG
#define	BARG SARG
#define	LLARG SARG
#define	LARG SARG
#define	ILIST SARG
#include "arg_vals.h"
#undef	SARG
};

static struct opts_comments opts_comments = {
#define	SARG(B, D, C) C,
#include "arg_vals.h"
#undef SARG
};
static struct opts_comments opts_names = {
#define	SARG(B, D, C) #B,
#include "arg_vals.h"
#undef SARG
};
static struct opts_comments opts_def_strings = {
#define	SARG(A, B, C) #B,
#undef ILIST
#define	ILIST(A, B, C) B,
#include "arg_vals.h"
#undef STARG
#undef SARG
#undef LLARG
#undef LARG
#undef CARG
#undef BARG
#undef IARG
#undef ILIST
#undef START_SECTION
#undef END_SECTION
#undef SET_OPENING_BLURB
};
static const char min_max_str[] = "# Minimum value 0, Maximum value";
static const char HASH[] = "# ";

static int
option_cmp(const void *a, const void *b)
{
	struct option_entry *as, *bs;
	as = (struct option_entry *)a;
	bs = (struct option_entry *)b;
	if (bs->len != as->len) {
		return (bs->len - as->len);
	}
	return (strncasecmp(as->key, bs->key, bs->len));
}

static int
check_and_free_option_store()
{
	int i;
	int err = 0;

	for (i = 0; i < opt_store.nentries; i++) {
		if (opt_store.entries[i].info.used == 0 &&
		    opt_store.entries[i].info.built_in_option == 0) {
			err++;
			if (opt_store.entries[i].info.filename != NULL) {
				fprintf(stderr,
				    "Unknown option \"%.*s\" at line"
				    " %d in file \"%s\"\n",
				    opt_store.entries[i].len,
				    opt_store.entries[i].key,
				    opt_store.entries[i].info.lineno.start,
				    opt_store.entries[i].info.filename);
			} else {
				fprintf(stderr, "Unknown option \"%.*s\" "
				    "at argument %d on command line\n",
				    opt_store.entries[i].len,
				    opt_store.entries[i].key,
				    opt_store.entries[i].info.lineno.end - 1);
			}
		}
		free(opt_store.entries[i].value);
	}
	free(opt_store.entries);
	opt_store.entries = NULL;
	opt_store.nentries = 0;
	return (err);
}

static int
store_option_with_val(char *key, char *value, long long *x,
	struct option_info *infop)
{
	struct option_entry e;
	struct option_entry *ne;

	if ((ne = realloc(opt_store.entries, (opt_store.nentries + 1) *
		sizeof (struct option_entry))) == NULL) {
		return (0);
	}

	opt_store.entries = ne;

	if ((e.value = strdup(value)) == NULL)
		return (0);
	e.key = key;
	e.len = strlen(key);
	e.info = *infop;
	e.info.no_report = 1;
	e.info.had_error = 0;
	e.info.used = 0;
	e.info.endchrs = NULL;

	ne = bsearch(&e, opt_store.entries, opt_store.nentries,
	    sizeof (struct option_entry), option_cmp);
	if (ne == NULL) {
		if (x == NULL) {
			e.lv = do_maths(0, e.value, NULL, &e.info);
		} else {
			e.lv = *x;
		}
		opt_store.entries[opt_store.nentries++] = e;
		qsort(opt_store.entries, opt_store.nentries,
		    sizeof (struct option_entry), option_cmp);
	} else {
		if (x == NULL) {
			e.lv = ne->lv;
		} else {
			e.lv = *x;
		}
		free(ne->value);
		*ne = e;
		ne->lv = do_maths(e.lv, e.value, NULL, &e.info);
	}

	return (1);
}
static int
store_option(char *key, char *value, struct option_info *infop)
{
	return (store_option_with_val(key, value, NULL, infop));
}

/*PRINTFLIKE2*/
static void
parse_error(struct option_info *infop, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	if (!infop) {
		vlog(LOG_WARNING, fmt, ap);
		return;
	}
	if (infop->no_report) {
		infop->had_error = 1;
		return;
	}
	if (infop->filename) {
		if (infop->lineno.start >= infop->lineno.end - 1) {
			plog(LOG_WARNING, gettext("Parse error: at line %d "
			    "of file \"%s\"\n"),
			    infop->lineno.start, infop->filename);
		} else {
			plog(LOG_WARNING,
			    gettext("Parse error: between line %d and %d "
			    "of file \"%s\"\n"),
			    infop->lineno.start, infop->lineno.end - 1,
			    infop->filename);
		}
		if (infop->line) {
			plog(LOG_WARNING, "Error processing option \"%s\"\n",
			    infop->line);
		}
	} else {
		plog(LOG_WARNING, "Parse error, on command line option "
		    "\"%s\"\n", infop->line);
	}
	vlog(LOG_WARNING, fmt, ap);
}
void
usage(const char *prog)
{
	const char *tmp = strrchr(prog, '/');
	if (tmp == NULL)
		tmp = prog;
	else
		tmp = tmp + 1;
	(void) fprintf(stderr,
	    gettext("USAGE: %s [ -f option_file ][ -o option=value]\n"), tmp);
	(void) fprintf(stderr,
	    gettext("For options type\n       %s -?\n"), tmp);
	exit(1);
}
/*
 * check_ends.
 *	returns true if c is one of the endchrs
 */
static int
check_ends(char c, struct option_info *infop)
{
	int x = (infop != NULL && infop->endchrs != NULL &&
	    strchr(infop->endchrs, c) != NULL);
	return (x);
}

static long long
my_strtoll(char *in, char **out, int base, struct option_info *infop)
{
	char *end;
	char *unused;
	long long temp;
	temp = strtoll(in, &end, base);
	if (end != NULL && !isspace(*end)) {
		if (*end == 'c' || *end == 'C') {
			end++;
		} else if (*end == 'b' || *end == 'B') {
			end++;
			temp *= 512;
		} else if (*end == 'k' || *end == 'K') {
			end++;
			temp *= 1024;
		} else if (*end == 'm' || *end == 'M') {
			end++;
			temp *= (1024 * 1024);
		} else if (*end == 'g' || *end == 'G') {
			end++;
			temp *= (1LL<<30);
		} else if (*end == 't' || *end == 'T') {
			end++;
			temp *= (1LL<<40);
		} else if (*end == 'p' || *end == 'P') {
			end++;
			temp *= (1LL<<50);
		} else if (*end == 'e' || *end == 'E') {
			end++;
			temp *= (1LL<<60);
		} else if (*end != NULL && !isspace(*end) &&
		    !check_ends(*end, infop) &&
		    choose_operator(end, &unused) == set) {
			parse_error(infop,
			    gettext("Expected one of [bkmgtpe]"
			    " a space or an operator not \"%c\"\n"),
			    *end++);
		}
	}
	if (out != NULL)
		*out = end;
	return (temp);
}
static int
has_white_space(const char *str)
{
	while (*str != NULL) {
		if (isspace(*(str++))) {
			return (1);
		}
	}
	return (0);
}

void
print_args(int argc, char **argv, void (*pprintf)(const char *fmt, ...))
{
	int i;
	char *str = NULL;
	int len = 0;

	for (i = 0; i < argc; i++) {
		char *tmp;
		char *quote;
		int olen = len;

		len += (strlen(argv[i]));

		if (has_white_space(argv[i])) {
			quote = "\"";
			len += 3;
		} else {
			quote = "";
			len += 1;
		}

		tmp = realloc(str, len + 1);
		if (tmp == NULL) {
			free(str);
			return;
		} else {
			str = tmp;
		}
		if (i == 0) {
			tmp = "";
			len--;
		} else {
			tmp = " ";
		}

		snprintf(&str[olen], len + 1 - olen, "%s%s%s%s", tmp, quote,
		    argv[i], quote);
	}
	pprintf("%s\n", str);
	free(str);
}

int
do_args(int argc, char **argv, void (*pprintf)(char *fmt, ...),
	const char *path)
{
	extern char *optarg;
	int c;

	opts_init();
	set_defaults();

	while ((c = getopt(argc, argv, "?f:o:h:d:")) != EOF) {
		switch (c) {
			case 'f':
				if (!read_config_file(optarg, pprintf, path))
					return (0);
				break;
			case 'o': {
				struct option_info info;

				info.filename = NULL;
				info.lineno.end  = optind;
				info.line = strdup(optarg);
				info.no_report = 0;
				(void) process_option(optarg, &info);
				if (info.line)
					free(info.line);
				break;
			}
			case 'h': /* print the html */
				print_html(*argv, optarg);
				break;
			case 'd': /* dump the current options */
				(void) print_file(*argv, optarg,
				    print_option_file);
				break;
			case '?':
				print_defaults(stdout);
				exit(0);
			default:
				usage(*argv);
		}
	}
	return (1);
}
static char *
get_next_line(char *line, char **ret, struct lineinfo *linenop)
{
	char *x, *y;
	char out = 0;
	char waswhite = 0;
	char comment = 0;
	for (y = x = line; *x != NULL && out == 0; ) {
		switch (*x) {
			case '\\':
				if (x == line)
					line++;
				x++;
				if (*x == NULL)
					out = 1;
				else {
					if (x == line)
						line++;
					if (*x == '\n')
						linenop->end += 1;
					x++;
				}
				break;
			case '#':
				x++;
				comment = 1;
				break;
			case '\n':
				x++;
				out = 1;
				waswhite = 0;
				linenop->end += 1;
				break;
			case ' ':
			case '\t':
				if (0 == comment && 0 == waswhite) {
					*y++ = *x++;
					waswhite = 1;
				} else
					x++;
				break;
			default:
				if (!comment)
					*y++ = *x++;
				else
					x++;
				waswhite = 0;
				break;
		}
	}
	*ret = x;
	*y = NULL;
	return (line);
}
static struct ilist
parse_comma_list(char *in, struct option_info *infop)
{
	char *x = in;
	long long temp, wtemp;
	int *res;
	struct ilist out;
	int count = 0;
	struct option_info info;


	out.wlen = 0;
	out.vals = NULL;
	out.weightings = NULL;
	if (infop) {
		info = *infop;
	} else {
		info.built_in_option = 1;
		info.filename = "(built in default)";
		info.line = NULL;
	}

	do {
		char *end;
		int i;

		info.endchrs = OPEN_BRACKET_STR;
		errno = 0;
		temp = do_maths(0, x, &end, &info);
		if (errno) {
			parse_error(&info,
			    gettext("error parsing \"%c\" in lint %s\n"),
			    *x, in);
		}
		if (temp == 0)
			break;
		if (end != NULL) {
			if (*end == OPEN_BRACKET) {
				x = end + 1;
				errno = 0;
				info.endchrs = CLOSE_BRACKET_STR;
				wtemp = do_maths(0, x, &end, &info);
					if (errno) {
						parse_error(&info,
						    gettext("error parsing "
						    "\"%c\" in lint %s\n"),
						    *x, in);
					}
				if (*end != CLOSE_BRACKET) {
					parse_error(&info, gettext(
					    "error parsing list %s\n"), in);
					exit(1);
				}
				end++;
				if (wtemp == 0)
					continue;
			} else {
				wtemp = 1;
			}
			i = out.wlen;
			out.wlen += wtemp;
			res = realloc(out.weightings, out.wlen *
			    sizeof (int));
			if (res == NULL) {
				REALLOC_ERROR(out.weightings, out.wlen *
				    sizeof (int));
				exit(1);
			}
			out.weightings = res;
			while (i < out.wlen) {
				out.weightings[i++] = count;
			}

			if (*end == COMMA)
				end++;
		}
		if (temp > UINT_MAX) {
			(void) fprintf(stderr, gettext(
			    "%lld (%#llx) greater than %d (%#x)\n"),
			    temp, temp, UINT_MAX, UINT_MAX);
			exit(1);
		}
		x = end;
		res = realloc(out.vals, ++count * sizeof (int));
		if (res == NULL) {
			REALLOC_ERROR(out.vals, count * sizeof (int));
			exit(1);
		}
		res[count-1] = (int)temp;
		out.vals = res;
	} while (temp != 0);

	if (infop) {
		const char *saveendchrs;
		saveendchrs = infop->endchrs;
		*infop  = info;
		infop->endchrs = saveendchrs;
	}
	return (out);
}
static void
usage_track_list(struct ilist *ilp)
{
	int i;
	int last = -1;
	int count = 0;
	char x[128];

	for (i = 0; i < ilp->wlen; i++) {
		if (ilp->weightings[i] != last) {
			if (count) {
				(void) sprintf(x, "(%d)", count);
				USAGE_TRACKING_STORE_VALUE(x);
			}
			(void) sprintf(x, "%s%#x", i ? ", " : "",
			    ilp->vals[ilp->weightings[i]]);
			USAGE_TRACKING_STORE_VALUE(x);
			last = ilp->weightings[i];
			count = 1;
		} else {
			count++;
		}
	}
	USAGE_TRACKING_STORE_VALUE("\n");
}

static void
fprint_ilist(FILE *out, struct ilist *ilp)
{
	int i;
	int last = -1;
	int count = 0;

	for (i = 0; i < ilp->wlen; i++) {
		if (ilp->weightings[i] != last) {
			if (count) {
				(void) fprintf(out, "(%d)", count);
			}
			(void) fprintf(out, "%s%#x", i ? ", " : "",
			    ilp->vals[ilp->weightings[i]]);
			last = ilp->weightings[i];
			count = 1;
		} else {
			count++;
		}
	}
	(void) fprintf(out, "\n");
}
static long long
add(long long x, long long y)
{
	return (x + y);
}
static long long
sub(long long x, long long y)
{
	return (x - y);
}
static long long
mul(long long x, long long y)
{
	return (x * y);
}
static long long
divide(long long x, long long y)
{
	return (x / y);
}
static long long
mod(long long x, long long y)
{
	return (x % y);
}
static long long
and(long long x, long long y)
{
	union {
		long long ll;
		uint_t i[sizeof (long long)/sizeof (uint_t)];
	} xu, yu, r;
	int i;

#ifdef __lint
	memset(&r, 0, sizeof (r));
#endif
	xu.ll = x;
	yu.ll = y;

	for (i = 0; i < sizeof (long long)/sizeof (uint_t); i++) {
		r.i[i] = xu.i[i] & yu.i[i];
	}
	return (r.ll);
}
static long long
or(long long x, long long y)
{
	union {
		long long ll;
		uint_t i[sizeof (long long)/sizeof (uint_t)];
	} xu, yu, r;
	int i;

	xu.ll = x;
	yu.ll = y;
#ifdef __lint
	memset(&r, 0, sizeof (r));
#endif

	for (i = 0; i < sizeof (long long)/sizeof (uint_t); i++) {
		r.i[i] = xu.i[i] | yu.i[i];
	}
	return (r.ll);
}
static long long
xor(long long x, long long y)
{
	union {
		long long ll;
		uint_t i[sizeof (long long)/sizeof (uint_t)];
	} xu, yu, r;
	int i;

#ifdef __lint
	memset(&r, 0, sizeof (r));
#endif
	xu.ll = x;
	yu.ll = y;

	for (i = 0; i < sizeof (long long)/sizeof (uint_t); i++) {
		r.i[i] = xu.i[i] ^ yu.i[i];
	}
	return (r.ll);
}
/*ARGSUSED*/
static long long
set(long long x, long long y)
{
	return (y);
}
static int
get_var(char *key, char **out, long long *v, struct option_info *infop)
{
#define	START_SECTION(A, B)
#define	END_SECTION(A)
#define	SET_OPENING_BLURB(A)
	char *x, *env;
	char c;
#define	STARG(B, D, C)
#define	LLARG(B, D, C) if (!strcasecmp(key, opts_names.B))  {\
	*out = key + strlen(opts_names.B);\
	*v = opts.B;\
	return (1);\
}
#define	ILIST(A, B, C)
#define	LARG(B, D, C) LLARG(B, D, C)
#define	IARG(B, D, C) LLARG(B, D, C)
#define	CARG(B, D, C) LLARG(B, D, C)
#define	BARG(B, D, C) LLARG(B, D, C)
#define	SARG(B, D, C) LLARG(B, D, C)
#include "arg_vals.h"
#undef STARG
#undef SARG
#undef CARG
#undef BARG
#undef IARG
#undef LLARG
#undef LARG
#undef ILIST
#undef START_SECTION
#undef END_SECTION
#undef SET_OPENING_BLURB
	x = key;
	while (!isspace(*x) && *x != NULL)
		++x;
	c = *x;
	*x = NULL;

	if (get_long_long_option(key, v) == OPT_FAILED &&
			(env = getenv(key)) != NULL) {
		*x = c;
		errno = 0;
		*v = my_strtoll(env, NULL, 0, infop);
		if (errno != 0)
			return (0);
		return (1);
	}
	return (1);
}
static int
do_variable(char *str, char **out, long long *ret, struct option_info *infop)
{
	char *line = str;
	while (isspace(*str))
		++str;

	if (*str == DOLLAR) {
		if (*(++str) == OPEN_BRACE) {
			char *x;

			str++;
			x = strchr(str, CLOSE_BRACE);
			if (x == NULL) {
				parse_error(infop, gettext(
				    "Error parsing variable: no close "
				    "brace: %s\n"), line);
				exit(1);
			}
			*x = NULL;
			if (get_var(str, out, ret, infop) == 0) {
				exit(1);
			}
			*x = CLOSE_BRACE;
			str = ++x;
		} else {
			if (get_var(str, out, ret, infop) == 0) {
				exit(1);
			}
			str = *out;
		}
		*out = str;
		return (1);
	}
	*out = str;
	return (0);
}
struct operators {
	char op;
	operator_t func;
};
static struct operators operators[] = {
	{ '+', add },
	{ '-', sub },
	{ '*', mul },
	{ '/', divide },
	{ '%', mod },
	{ '&', and },
	{ '|', or },
	{ '^', xor },
	{ '=', set },
	{ ' ', NULL }
};

static long long
(*choose_operator(char *str, char **newstr))(long long x, long long y)
{
	struct operators *ops;
	while (isspace(*str))
		++str;

	for (ops = &operators[0]; ops->func != NULL; ops++) {
		if (ops->op == *str) {
			*newstr = str + 1;
			return (ops->func);
		}
	}
	*newstr = str;
	return (set);
}
static long long
do_maths(long long cur, char *str, char **nstr, struct option_info *infop)
{
	long long (*op)(long long, long long);
	long long x;
	int v;
	char *newstr;
	char *line = str;

	while (isspace(*str))
		++str;
	if (do_variable(str, &newstr, &x, infop))
		cur = x;
	str = newstr;
	op = choose_operator(str, &newstr);
	str = newstr;
	while (isspace(*str))
		++str;
	if (*str == OPEN_BRACKET) {
		x = do_maths(0, ++str, &newstr, infop);
		while (isspace(*str))
			++str;
		if (*str == CLOSE_BRACKET)
			++str;
	} else if (*str == CLOSE_BRACKET) {
		if (nstr)
			*nstr = str;
		return (cur);
	} else if (!(v = do_variable(str, &newstr, &x, infop)) &&
	    strchr("0123456789", *str) == NULL &&
	    (infop == NULL || infop->endchrs == NULL ||
	    strchr(infop->endchrs, *str) == NULL)) {
		parse_error(infop, gettext("error at char \"%c\" "
		    "expected one of [0-9] or (\n"), *str);
		if (nstr)
			*nstr = str;
		return (cur);
	} else if (*str == NULL) {
		if (nstr)
			*nstr = str;
		return (cur);
	} else if (!v) {
		errno = 0;
		x = my_strtoll(str, &newstr, 0, infop);
		if (errno) {
			parse_error(infop, gettext(
			    "error Bad number at \"%c\" in %s\n"),
			    *str, line);
			if (nstr)
				*nstr = newstr;
			return (cur);
		}
	}
	str = newstr;
	while (isspace(*str))
		++str;
	if (*str == NULL || *str == CLOSE_BRACKET || check_ends(*str, infop)) {
		if (nstr)
			*nstr = str;
		cur = op(cur, x);
		if (*str != NULL && !check_ends(*str, infop)) {
			str++;
			return (do_maths(cur, str, nstr, infop));
		} else {
			return (cur);
		}
	} else {
		return (do_maths(op(cur, x), str, nstr, infop));
	}
}
static int
do_key_value(char *key, char *value, struct option_info *infop)
{
/* STARG String ARG */
#define	START_SECTION(A, B)
#define	END_SECTION(A)
#define	SET_OPENING_BLURB(A)
#ifdef NOT
#define	STARG(B, D, C) if (!strcasecmp(key, opts_names.B))  {\
	NOT_NULL_FREE(opts.B);\
	if (strcmp(value, "NULL") == 0) \
		opts.B = NULL; \
	else if ((opts.B = strdup(value)) == NULL) \
		return (0); \
	else \
		return (1); \
}
#else
#define	STARG(B, D, C) if (!strcasecmp(key, opts_names.B))  {\
	if (strcmp(value, "NULL") == 0) \
		opts.B = NULL; \
	else {\
		opts.B = value; \
		return (1); \
	} \
}
#endif
#define	NARROW_ARG(A, B, C) if (!strcasecmp(key, opts_names.C)) {\
	long x; \
	x = do_maths(opts.C, value, NULL, infop);\
	if (x > B) \
		opts.C = B; \
	else \
		opts.C = (A) x; \
	return (1); \
}
#define	LLARG(B, D, C) if (!strcasecmp(key, opts_names.B))  {\
	opts.B = do_maths(opts.B, value, NULL, infop);\
	return (1); \
}
#define	ILIST(A, B, C) if (!strcasecmp(key, opts_names.A))  {\
		opts.A = parse_comma_list(value, infop);\
		return (1); \
	}
#define	LARG(B, D, C) NARROW_ARG(ulong_t, ULONG_MAX, B)
#define	IARG(B, D, C) NARROW_ARG(uint_t, UINT_MAX, B)
#define	CARG(B, D, C) NARROW_ARG(uchar_t, UCHAR_MAX, B)
#define	BARG(B, D, C) NARROW_ARG(uchar_t, 1, B)
#define	SARG(B, D, C) NARROW_ARG(ushort_t, USHRT_MAX, B)
#include "arg_vals.h"
#undef STARG
#undef SARG
#undef CARG
#undef BARG
#undef IARG
#undef LLARG
#undef LARG
#undef ILIST
#undef START_SECTION
#undef END_SECTION
#undef SET_OPENING_BLURB
	return (0);
}
static int
process_option(char *option, struct option_info *infop)
{
	char *value;
	char c;
	int x;

	if (*option == '#' || (value = strchr(option, '=')) == NULL)
		return (0);

	c = *value;
	*value = NULL;
	/* We have option pointing to the option and value to the value */
	x = do_key_value(option, value+1, infop);
	if (x == 0) {
		infop->built_in_option = 0;
		(void) store_option(option, value+1, infop);
		*value = c;
		(void) putenv(option);
	} else {
		infop->built_in_option = 1;
		(void) store_option(option, value+1, infop);
		*value = c;
	}
	return (1);
}

#define	START_SECTION(A, B)
#define	END_SECTION(A)
#define	SET_OPENING_BLURB(A)
static void
set_defaults(void)
{
	struct option_info info;
	long long x;
	info.built_in_option = 1;
	info.filename = "(built in default)";
	info.line = NULL;
	info.endchrs = NULL;


#define	STARG(B, D, C) opts.B = (D == NULL ? NULL : strdup(D)); \
	(void) store_option(#B, \
	(D == NULL ? "NULL" : strdup(D)), &info);
#define	SARG(B, D, C) x = opts.B = D;\
	(void) store_option_with_val(#B, #D, &x, &info);
#define	IARG(B, D, C) SARG(B, D, C)
#define	CARG(B, D, C) SARG(B, D, C)
#define	BARG(B, D, C) SARG(B, D, C)
#define	LLARG(B, D, C) SARG(B, D, C)
#define	LARG(B, D, C) SARG(B, D, C)
#define	ILIST(A, B, C) opts.A = parse_comma_list(B, NULL);
#include "arg_vals.h"
#undef	STARG
#undef	SARG
#undef	IARG
#undef	CARG
#undef	BARG
#undef	LLARG
#undef	LARG
#undef	ILIST
}
#undef START_SECTION
#undef END_SECTION
#undef SET_OPENING_BLURB
static const char *
search_back(const char *start, const char *from, const char *str, int len)
{
	while (from >= start) {
		int i;

		for (i = 0; i < len; i++)
			if (str[i] == *from)
				return (from);
		from--;
	}
	return (NULL);
}
static char *
strip_html(const char *str)
{
	char *tmp, *ret;
	const char *in;
	int depth = 0;
	if ((ret = strdup(str)) == NULL)
		return (ret);
	in = str;

	tmp = ret;
	while (*str != NULL) {
		if (*str == '<') {
			depth++;
			str++;
		} else if (*str == '>') {
			if (--depth < 0) {
				(void) fprintf(stderr, gettext(
				    "Unbalanced html tags\n\n%s\n"), in);
			}
			str++;
		} else if (depth == 0)
			*tmp++ = *str++;
		else
			str++;
	}
	*tmp = NULL;
	return (ret);
}
static const char *
search_forward(const char *str, char c, int len)
{
	int x;

	for (x = 0; x < len; x++) {
		if (str[x] == NULL)
			return (NULL);
		if (str[x] == c)
			return (&str[x]);
	}
	return (NULL);
}


static void
print_line(FILE *f, const char *line, const char *newline, const char *endl)
{
	const char *this_line;
	const char *x;

	for (this_line = line; /* until break */; /* nothing */) {
		int len;
		if ((x = search_forward(this_line, '\n', LINE_LEN)) == NULL) {
			if (strlen(this_line) < LINE_LEN) {
				(void) fprintf(f, "%s%s\n", newline,
				    this_line);
				break;
			} else {
				x = search_back(this_line,
				    &this_line[LINE_LEN - 1], " \t", 2);
			}
		}
		if (x != NULL) {
			/*LINTED*/
			len = x - this_line;
		} else
			len = (int)strlen(this_line);
		(void) fprintf(f, "%s%.*s", newline, len, this_line);

		if (x == NULL) {
			(void) fprintf(f, "\n");
			break;
		} else {
			(void) fprintf(f, "%s", endl == NULL ? "\n" : endl);
			this_line = x + 1;
		}
	}
}
static void
print_comment(FILE *f, const char *newline,
	const char *option,  const char *str)
{
	char *to_free;

	if ((to_free = strip_html(str)) == NULL)
		return;

	if (option != NULL)
		(void) fprintf(f, "%s\n%s%s\n", newline, newline, option);
	print_line(f, to_free, newline, NULL);
	free(to_free);
}
static void
print_copyright(FILE *f, const char *newline, const char *endline)
{
	(void) fprintf(f, "%s%s\n%s\tDiskomizer %s%s\n%s\t"
	    "Copyright Sun Microsystems Inc 1988-%s%s\n%s%s\n",
	    newline, endline, newline, VERSION, endline,
	    newline, THIS_YEAR, endline, newline, endline);
}
static void
print_max_min_disclaimer(FILE *f, const char *newline, char *str)
{
	print_comment(f, newline, str, gettext(
	    "The maximum and minimum values listed are the values at which "
	    "the options are capped. They do not mean that "
	    "the program will run with the values given and should not"
	    " be taken to be recommended values."));
	(void) fprintf(f, "%s\n", newline);
}
int
print_options(FILE *f)
{
	char x[256];
	int i;
	char buf[1024];
	struct opts_comments *opts_namesp;

#define	UPPER(A) { \
	for (i = 0; A[i] != NULL; i++) { \
		if (islower(A[i])) { \
			x[i] = toupper(A[i]); \
		} else { \
			x[i] = A[i]; \
		} \
	} \
	x[i] = NULL; \
}


	print_copyright(f, HASH, "");
	print_comment(f, HASH, NULL, OPENING_STR);
	print_max_min_disclaimer(f, HASH, "Option file");
	(void) fprintf(f, "# %s=%s\n", gettext("Option Name"),
		gettext("Value"));
	(void) fprintf(f, "# %s=%s\n", "-----------", "-----");
#define	START_SECTION(A, B) UPPER(opts_snames.A); \
	print_comment(f, HASH, x, opts_scomments.A);
#define	END_SECTION(A)
#define	SET_OPENING_BLURB(A)
#define	ILIST(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, x, gettext(opts_comments.A)); \
	(void) fprintf(f, "%s=", x); \
	fprint_ilist(f, &opts.A);
#define	IARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, x, gettext(opts_comments.A)); \
	(void) fprintf(f, "%s %u\n", min_max_str, UINT_MAX); \
	(void) sprintf(buf, gettext("Default is \"%s\" = %u  (%#x)"), \
		opts_def_strings.A, B, B); \
	print_comment(f, HASH, NULL, &buf[0]); \
	(void) fprintf(f, "%s=%u # %#x\n", x, opts.A, opts.A);
#define	CARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, x, gettext(opts_comments.A)); \
	(void) fprintf(f, "%s %u\n", min_max_str, \
		(int)UCHAR_MAX); \
	(void) sprintf(buf, gettext("Default is \"%s\" = %u  (%#x)"), \
		opts_def_strings.A, B, B); \
	print_comment(f, HASH, NULL, &buf[0]); \
	(void) fprintf(f, "%s=%d # %#2.x\n", x, opts.A, opts.A);
#define	BARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, x, gettext(opts_comments.A)); \
	(void) fprintf(f, "%s %u\n", min_max_str, 1); \
	(void) sprintf(buf, gettext("Default is \"%s\" = %u  (%#x)"), \
		opts_def_strings.A, B, B); \
	print_comment(f, HASH, NULL, &buf[0]); \
	(void) fprintf(f, "%s=%d\n", x, opts.A);
#define	STARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, x, gettext(opts_comments.A)); \
	(void) fprintf(f, "%s=", x); \
	print_line(f, opts.A == NULL ? "NULL" : opts.A, "", " \\\n\t");
#define	SARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, x, gettext(opts_comments.A)); \
	(void) fprintf(f, "%s %u\n", min_max_str, \
		USHRT_MAX); \
	(void) fprintf(f, "%s=%d # %#4.x\n", x, opts.A, opts.A);
#define	LLARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, x, gettext(opts_comments.A)); \
	(void) fprintf(f, "%s %llu\n",  min_max_str,\
		LONGLONG_MAX); \
	(void) sprintf(buf, gettext("Default is \"%s\" = %llu  (%#llx)"), \
		opts_def_strings.A, B, B); \
	print_comment(f, HASH, NULL, &buf[0]); \
	(void) fprintf(f, "%s=%llu # %#llx\n", x, opts.A, opts.A);
#define	LARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, x, gettext(opts_comments.A)); \
	(void) fprintf(f, "%s %lu\n",  min_max_str,\
		ULONG_MAX); \
	(void) sprintf(buf, gettext("Default is \"%s\" = %lu  (%#lx)"), \
		opts_def_strings.A, B, B); \
	print_comment(f, HASH, NULL, &buf[0]); \
	(void) fprintf(f, "%s=%lu # %#lx\n", x, opts.A, opts.A);
#include "arg_vals.h"
#undef UPPER
#undef STARG
#undef ILIST
#undef SARG
#undef LLARG
#undef LARG
#undef CARG
#undef BARG
#undef IARG
#undef START_SECTION
#undef END_SECTION
#undef SET_OPENING_BLURB
	print_all_option_entries(f);
	return (1);
}
int
save_options(const char *file)
{
	return (print_file(NULL, file, print_option_file));
}

/*ARGSUSED*/
static int
print_option_file(const char *prog, const char *file, FILE *f)
{
	return (print_options(f));
}
/*ARGSUSED*/
static int
print_html_file(const char *prog, const char *filename, FILE *f)
{
#define	LIST_STRING "<LI><B><A NAME=\"%s\">%s</A></B>\n<BR>\n"
#define	TITLE_STRING "<LI><H3><A NAME=\"%s\">%s</A></H3>\n"
#define	CONTENT_STRING "%s\n<P>\n<OL>\n"
/*
 * lint will check the printf like format strings, but only if they are
 * done as macros and not if as constant strings. Hence this
 * giggery pokery. Constant strings take up less space in the binary.
 */
#ifndef __lint
	const char LI_STR[] = LIST_STRING;
	const char TITLE_STR[] = TITLE_STRING;
	const char CONTENT_STR[] = CONTENT_STRING;
#else
#define	LI_STR LIST_STRING
#define	TITLE_STR TITLE_STRING
#define	CONTENT_STR CONTENT_STRING
#endif
	char x[256];
	char *now;
	int i;

	now = alloc_time_now_fmt("%Y" "%m" "%d:%H" "%M" "%S00");


	(void) fprintf(f, "<HTML>\n<!--START_PAGE_CONFIG\n"
	    "page.sidebar=off\n"
	    "page.greybar=on\n"
	    "page.greybar.entry.1=Overview|\"diskomizer.html\"\n"
	    "page.greybar.entry.2="
	    "Getting Started|\"Getting_started.html\"\n"
	    "page.greybar.entry.3=Options %s|\"Options-%s.html\"\n"
	    "page.greybar.entry.4=Options %s|\"Options-%s.html\"\n"
	    "page.greybar.entry.5=Options %s|\"Options-%s.html\"\n"
	    "page.greybar.entry.6=New Features|\"new-features.html\"\n"
	    "page.greybar.entry.7=Interpreting Errors|"
	    "\"Interpreting_Errors.html\"\n"
	    "END_PAGE_CONFIG-->\n",
#ifdef __sparc
	    "i86pc", "i86pc",
	    "AMD 64", "amd64",
	    LONG_BIT == 32 ? "SPARC v9" : "SPARC v7",
	    LONG_BIT == 32 ? "64" : "32"
#else
	    "SPARC v7", "32",
	    "SPARC v9", "64",
	    LONG_BIT == 32 ? "AMD 64" : "i86pc",
	    LONG_BIT == 32 ? "amd64" : "i86pc"
#endif
	    /* for cystye */);
	(void) fprintf(f,
	    "<HEAD>\n\t<TITLE>Diskomizer Options</TITLE>\n"
	    "\t<META NAME=\"GENERATOR\" CONTENT=\"diskomizer %s\">\n"
	    "\t<META NAME=\"CREATED\" CONTENT=\"%s\">\n"
	    "\t<META NAME=\"CHANGEDBY\" CONTENT=\" \">\n"
	    "\t<META NAME=\"CHANGED\" CONTENT=\"%s\">\n"
	    "\t<META NAME=\"KEYWORDS\" CONTENT="
	    "\"diskomizer disk testing stress test\">\n"
	    "\t<META HTTP-EQUIV=\"Content-Type\" "
	    /* CSTYLED */
	    "CONTENT=\"text/html; charset=iso-8859-1\">\n",
	    VERSION, now, now);
	free(now);
	(void) fprintf(f, "</HEAD>\n<BODY>\n");
	print_copyright(f, "<! ", ">");
	(void) fprintf(f, "<H1>Diskomizer %s</H1><P>\n", VERSION);
	(void) fprintf(f, "%s\n<P>\n", OPENING_STR);
	print_max_min_disclaimer(f, "", NULL);
	(void) fprintf(f, "<HR>\n");
	(void) fprintf(f, "<OL>\n");
#define	LIST_ITEM(A) (void) fprintf(f, LI_STR, A, A)

#define	UPPER(A) { \
	for (i = 0; A[i] != NULL; i++) { \
		if (islower(A[i])) { \
			x[i] = toupper(A[i]); \
		} else { \
			x[i] = A[i]; \
		} \
	} \
	x[i] = NULL; \
}

#define	START_SECTION(A, B) UPPER(opts_snames.A); \
	(void) fprintf(f, TITLE_STR, opts_snames.A, opts_snames.A); \
	if (opts_scomments.A != NULL) \
		(void) fprintf(f, CONTENT_STR, opts_scomments.A); \
	else \
		(void) fprintf(f, "<OL>");
#define	END_SECTION(A) (void) fprintf(f, "</OL>");
#define	SET_OPENING_BLURB(A)
#define	ILIST(A, B, C) UPPER(opts_names.A); \
	LIST_ITEM(x); \
	(void) fprintf(f, "%s", opts_comments.A); \
	(void) fprintf(f, "<BR>\n%s %u\n", &min_max_str[2], UINT_MAX); \
	(void) fprintf(f, "<BR>Default is \"%s\"", opts_def_strings.A); \
	(void) fprintf(f, "<BR>%s=%s\n", x, B);
#define	IARG(A, B, C) UPPER(opts_names.A); \
	LIST_ITEM(x); \
	(void) fprintf(f, "%s", opts_comments.A); \
	(void) fprintf(f, "<BR>%s %u\n", &min_max_str[2], UINT_MAX); \
	(void) fprintf(f, "<BR>Default is \"%s\"  = %u (%#x)", \
		opts_def_strings.A, B, B); \
	(void) fprintf(f, "<BR>%s=%u\n", x, B);
#define	CARG(A, B, C) UPPER(opts_names.A); \
	LIST_ITEM(x); \
	(void) fprintf(f, "%s", opts_comments.A); \
	(void) fprintf(f, "<BR>%s %u\n", &min_max_str[2], \
		(int)UCHAR_MAX); \
	(void) fprintf(f, "<BR>Default is \"%s\"  = %u (%#x)", \
		opts_def_strings.A, B, B); \
	(void) fprintf(f, "<BR>%s=%d\n", x, B);
#define	BARG(A, B, C) UPPER(opts_names.A); \
	LIST_ITEM(x); \
	(void) fprintf(f, "%s", opts_comments.A); \
	(void) fprintf(f, "<BR>%s %u\n", &min_max_str[2], 1); \
	(void) fprintf(f, "<BR>Default is \"%s\"  = %u (%#x)", \
		opts_def_strings.A, B, B); \
	(void) fprintf(f, "<BR>%s=%d\n", x, B);
#define	STARG(A, B, C) UPPER(opts_names.A); \
	LIST_ITEM(x); \
	(void) fprintf(f, "%s\n", opts_comments.A); \
	(void) fprintf(f, "<BR>Default is %s", opts_def_strings.A); \
	(void) fprintf(f, "<BR>%s=%s\n", x, B == NULL? "NULL" : B);
#define	SARG(A, B, C) UPPER(opts_names.A); \
	LIST_ITEM(x); \
	(void) fprintf(f, "%s", opts_comments.A); \
	(void) fprintf(f, "<BR>%s %u\n", &min_max_str[2], \
		USHRT_MAX); \
	(void) fprintf(f, "<BR>Default is \"%s\" = %u (%#x)", \
		opts_def_strings.A, B, B); \
	(void) fprintf(f, "<BR>%s=%s\n", x, opts_names.A);
#define	LLARG(A, B, C) UPPER(opts_names.A); \
	LIST_ITEM(x); \
	(void) fprintf(f, "%s", opts_comments.A); \
	(void) fprintf(f, "<BR>%s %llu\n", &min_max_str[2], \
		LONGLONG_MAX); \
	(void) fprintf(f, "<BR>Default is \"%s\" = %llu (%#llx)", \
		opts_def_strings.A, B, B); \
	(void) fprintf(f, "<BR>%s=%llu\n", x, B);
#define	LARG(A, B, C) UPPER(opts_names.A); \
	LIST_ITEM(x); \
	(void) fprintf(f, "%s", opts_comments.A); \
	(void) fprintf(f, "<BR>%s %lu\n",  &min_max_str[2], \
		ULONG_MAX); \
	(void) fprintf(f, "<BR>Default is \"%s\" = %lu (%#lx)", \
		opts_def_strings.A, B, B); \
	(void) fprintf(f, "<BR>%s=%lu\n", x, B);
#include "arg_vals.h"
(void) fprintf(f, "</OL>\n<HR>\n");
print_copyright(f, "<BR>", "");
(void) fprintf(f, "</BODY>\n</HTML>\n");
#undef UPPER
#undef ILIST
#undef STARG
#undef SARG
#undef LLARG
#undef LARG
#undef CARG
#undef BARG
#undef IARG
#undef START_SECTION
#undef END_SECTION
#undef SET_OPENING_BLURB
	return (1);
}

static void
report_back_opts()
{
	char *section;
	char x[256];
	int i;
	static const char option_str[] = "options";

	USAGE_TRACKING_OPEN_KEY(option_str, NULL, NULL);
#define	UPPER(A) { \
	for (i = 0; A[i] != NULL; i++) { \
		if (islower(A[i])) { \
			x[i] = toupper(A[i]); \
		} else { \
			x[i] = A[i]; \
		} \
	} \
	x[i] = NULL; \
}

#define	START_SECTION(A, B) section = opts_snames.A; \
	UPPER(section); \
	USAGE_TRACKING_OPEN_KEY(section, NULL, NULL);

#define	END_SECTION(A) USAGE_TRACKING_CLOSE_KEY();
#define	SET_OPENING_BLURB(A)
#define	STARG(A, B, C) \
		UPPER(opts_names.A); \
		USAGE_TRACKING_STORE_KEY_VALUE(x, NIL_STR(opts.A));
#define	ILIST(A, B, C) \
		UPPER(opts_names.A); \
		USAGE_TRACKING_OPEN_KEY(x, NULL, NULL); \
		usage_track_list(&opts.A); \
		USAGE_TRACKING_CLOSE_KEY();
#define	IARG(A, B, C) \
		UPPER(opts_names.A); \
		USAGE_TRACKING_STORE_KEY_VALUE_INT(x, opts.A);
#define	CARG(A, B, C) IARG(A, B, C)
#define	BARG(A, B, C) IARG(A, B, C)
#define	SARG(A, B, C) IARG(A, B, C)
#define	LLARG(A, B, C) IARG(A, B, C)
#define	LARG(A, B, C) IARG(A, B, C)

#include "arg_vals.h"

#undef UPPER
#undef ILIST
#undef STARG
#undef SARG
#undef LLARG
#undef LARG
#undef CARG
#undef BARG
#undef IARG
#undef START_SECTION
#undef END_SECTION
#undef SET_OPENING_BLURB

	USAGE_TRACKING_CLOSE_KEY();
}

static int
print_file(const char *prog, const char *filename,
	int (*print_func)(const char *, const char *, FILE *))
{
	if (strcmp(filename, "-")) {
		FILE *f;
		if ((f = fopen(filename, "w")) == NULL) {
			FOPEN_ERROR(filename, "w");
			return (0);
		}
		print_func(prog, filename, f);
		(void) fflush(f);
		if (fclose(f) != 0) {
			CLOSE_ERROR(fileno(f), filename);
			return (0);
		}
	} else {
		print_func(prog, filename, stdout);
		(void) fflush(stdout);
	}
	return (1);
}
static void
print_html(const char *prog, const char *filename)
{
	(void) print_file(prog, filename, print_html_file);
	exit(0);
}

static void
print_defaults(FILE *f)
{
	char buf[1024];
	int i;

	print_max_min_disclaimer(f, HASH, NULL);
	(void) fprintf(f, "# %s=%s\n", "Option Name", "Default Value");
	(void) fprintf(f, "# %s=%s\n", "-----------", "-------------");

#define	UPPER(A) { \
	for (i = 0; A[i] != NULL; i++) { \
		if (islower(A[i])) { \
			buf[i] = toupper(A[i]); \
		} else { \
			buf[i] = A[i]; \
		} \
	} \
	buf[i] = NULL; \
}

#define	START_SECTION(A, B) UPPER(opts_snames.A); \
	print_comment(f, HASH, buf, opts_scomments.A);
#define	END_SECTION(A)
#define	SET_OPENING_BLURB(A)
#define	ILIST(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, buf, opts_comments.A); \
	(void) fprintf(f, "%s %u\n", min_max_str, UINT_MAX); \
	(void) fprintf(f, "%s=%s\n", buf, B);
#define	IARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, buf, opts_comments.A); \
	(void) fprintf(f, "%s %u\n", min_max_str, UINT_MAX); \
	(void) sprintf(buf, "Default is \"%s\" = %u  (%#x)", \
		opts_def_strings.A, B, B); \
	print_comment(f, HASH, NULL, &buf[0]); \
	(void) fprintf(f, "%s=%u\n", buf, B);
#define	CARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, buf, opts_comments.A); \
	(void) fprintf(f, "%s %u\n", min_max_str, \
		(int)UCHAR_MAX); \
	(void) fprintf(f, "%s=%d\n", buf, B);
#define	BARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, buf, opts_comments.A); \
	(void) fprintf(f, "%s %u\n", min_max_str, 1); \
	(void) fprintf(f, "%s=%d\n", buf, B);
#define	STARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, buf, opts_comments.A); \
	(void) fprintf(f, "%s=%s\n", buf, B == NULL? "NULL" : B);
#define	SARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, buf, opts_comments.A); \
	(void) fprintf(f, "%s %u\n", min_max_str, \
		USHRT_MAX); \
	(void) fprintf(f, "%s=%s\n", buf, opts_names.A);
#define	LLARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, buf, opts_comments.A); \
	(void) fprintf(f, "%s %llx\n", min_max_str, \
		LONGLONG_MAX); \
	(void) sprintf(buf, "Default is \"%s\" = %llu  (%#llx)", \
		opts_def_strings.A, B, B); \
	print_comment(f, HASH, NULL, &buf[0]); \
	(void) fprintf(f, "%s=%llu\n", buf, B);
#define	LARG(A, B, C) UPPER(opts_names.A); \
	print_comment(f, HASH, buf, opts_comments.A); \
	(void) fprintf(f, "%s %lu\n",  min_max_str, \
		ULONG_MAX); \
	(void) sprintf(buf, "Default is \"%s\" = %lu (%#lx)", \
		opts_def_strings.A, B, B); \
	print_comment(f, HASH, NULL, &buf[0]); \
	(void) fprintf(f, "%s=%lu\n", buf, B);
#include "arg_vals.h"
#undef UPPER
#undef ILIST
#undef STARG
#undef SARG
#undef LLARG
#undef LARG
#undef CARG
#undef BARG
#undef IARG
#undef START_SECTION
#undef END_SECTION
#undef SET_OPENING_BLURB
}
int
read_config_file(char *file, void (*pprintf)(char *fmt, ...), const char *path)
{
	time_t last_read = 0;
	size_t len;
	char *ptr, *start;
	char *this, *next;
	int count = 0;
	struct option_info info;

	if ((ptr = map_file(file, &last_read, &len, pprintf, path)) == NULL) {
		pprintf("Unable to map option file: %s\n", file);
		return (0);
	}
	start = ptr;
	if (*(ptr+len-1) != '\n') {
		pprintf("Incomplete last line %s\n", file);
		(void) munmap(ptr, len);
		return (0);
	}
	*(ptr+len-1) = NULL;

	info.filename = file;
	info.lineno.end  = 1;
	info.no_report = 0;
	info.endchrs = "";
	while (ptr != NULL && *ptr != NULL) {
		info.lineno.start  = info.lineno.end;
		this = get_next_line(ptr, &next, &info.lineno);

		info.line = strdup(this);
		if (*this != NULL && process_option(this, &info))
			count++;
		if (info.line != NULL)
			free(info.line);
		ptr = next;
	}
	(void) mprotect(start, len, PROT_READ);
	return (count);
}
/*
 * Map a file, if the last_readp points to a time that is less
 * than the last modify time of the file.
 */
char *
map_file(char *file, time_t *last_readp, size_t *lenp,
	void (*pprintf)(char *fmt, ...), const char *path)
{
	struct stat buf;
	char *ptr;
	char *to_free = NULL;
	int fd;

	if (path == NULL || strchr(file, '/') != NULL) {
		if ((fd = open(file, O_RDONLY)) == -1) {
			OPEN_ERROR(file, O_RDONLY);
			return (NULL);
		}
	} else {
		char *tmp;
		char *opaque;
		char *x;
		x = strdup(path);
		tmp = x;
		/*
		 * if the diskomizer path starts with a colon then
		 * start by opening the current directory
		 */
		if (*tmp == ':') {
			fd = open(file, O_RDONLY);
		} else {
			fd = -1;
		}
		while (fd == -1 &&
		    (tmp = strtok_r(tmp, ":", &opaque)) != NULL) {
			to_free = realloc(to_free,
			    strlen(tmp) + strlen(file) + 2);
			(void) sprintf(to_free, "%s%s%s",
			    tmp, *tmp == NULL? "" : "/", file);

			fd = open(to_free, O_RDONLY);
			tmp = NULL;
		}
		free(x);
		if (fd == -1) {
			free(to_free);
			return (NULL);
		}
		file = to_free;
	}
	if (fstat(fd, &buf) == -1) {
		pprintf("Can't fstat %s: %s\n", file, strerror(errno));
		(void) close(fd);
		free(to_free);
		return (NULL);
	}

	if ((buf.st_mode & S_IFDIR) != 0 || buf.st_mtime <= *last_readp ||
	    buf.st_size == 0) {
		(void) close(fd);
		free(to_free);
		return (NULL);
	}

	if ((ptr = mmap(0, buf.st_size, PROT_READ|PROT_WRITE, MAP_PRIVATE,
	    fd, 0)) == MAP_FAILED) {
		pprintf("Can't mmap %s: %s\n", file, strerror(errno));
		(void) close(fd);
		free(to_free);
		return (NULL);
	}
	(void) close(fd);
	/*LINTED*/
	*lenp = buf.st_size;
	*last_readp = buf.st_mtime;
	if (to_free != NULL)
		free(to_free);
	return (ptr);
}
static void
print_all_option_entries(FILE *f)
{
	int i;

	fprintf(f, gettext(
	    "#\n# External options\n#These options were defined but "
	    "not understood by diskomizer, they could be for use "
	    "with shared libraries.\n#\n"));

	for (i = 0; i < opt_store.nentries; i++) {
		if (opt_store.entries[i].info.built_in_option == 0) {
			fprintf(f, "%.*s=%s\n", opt_store.entries[i].len,
			    opt_store.entries[i].key,
			    opt_store.entries[i].value);
		}
	}
}
#define	OVERFLOW_WARNING(A, B, K) \
	if ((A)->lv > B) {\
		plog(LOG_WARNING, \
			gettext("%s(%lld) overflows %d at %s:%d\n"), \
			K, (A)->lv, B, \
				NULL == (A)->info.filename ? \
				"(Command line)" : (A)->info.filename, \
				(A)->info.lineno.end); \
	}
static struct option_entry *
get_option_entry(const char *key)
{
	struct option_entry e;
	struct option_entry *r;

	e.key = key;
	e.len = strlen(key);

	if (opt_store.nentries) {
		r = ((struct option_entry *)
		    bsearch(&e, opt_store.entries, opt_store.nentries,
		    sizeof (struct option_entry), option_cmp));
		if (r != NULL) {
			r->info.used = 1;
		}
		return (r);
	} else {
		return (NULL);
	}
}
static option_status_t
get_char_option(const char *key, char *out)
{
	struct option_entry *e;

	if ((e = get_option_entry(key)) == NULL)
		return (OPT_FAILED);
	if (e->info.had_error) {
		*out = (char)do_maths(0, e->value, NULL, &e->info);
		return (OPT_FAILED);
	} else {
		*out = (char)e->lv;
		OVERFLOW_WARNING(e, CHAR_MAX, key);
	}
	return (OPT_OK);

}
static option_status_t
get_bool_option(const char *key, char *out)
{
	return (get_char_option(key, out));
}

static option_status_t
get_short_option(const char *key, short *out)
{
	struct option_entry *e;

	if ((e = get_option_entry(key)) == NULL)
		return (OPT_FAILED);
	if (e->info.had_error) {
		*out = (short)do_maths(0, e->value, NULL, &e->info);
		return (OPT_FAILED);
	} else {
		*out = (short)e->lv;
		OVERFLOW_WARNING(e, SHRT_MAX, key);
	}
	return (OPT_OK);

}
static option_status_t
get_int_option(const char *key, int *out)
{
	struct option_entry *e;

	if ((e = get_option_entry(key)) == NULL)
		return (OPT_FAILED);
	if (e->info.had_error) {
		*out = (int)do_maths(0, e->value, NULL, &e->info);
		return (OPT_FAILED);
	} else {
		*out = (int)e->lv;
		OVERFLOW_WARNING(e, INT_MAX, key);
	}
	return (OPT_OK);
}
static option_status_t
get_long_option(const char *key, long *out)
{
	struct option_entry *e;

	if ((e = get_option_entry(key)) == NULL)
		return (OPT_FAILED);
	if (e->info.had_error) {
		*out = (long)do_maths(0, e->value, NULL, &e->info);
		return (OPT_FAILED);
	} else {
		*out = (long)e->lv;
		OVERFLOW_WARNING(e, LONG_MAX, key);
	}
	return (OPT_OK);
}
static option_status_t
get_long_long_option(const char *key, long long *out)
{
	struct option_entry *e;

	if ((e = get_option_entry(key)) == NULL)
		return (OPT_FAILED);
	*out = (long long)do_maths(0, e->value, NULL, &e->info);
	return (OPT_OK);
}
static option_status_t
get_str_option(const char *key, char **out)
{
	struct option_entry *e;

	if ((e = get_option_entry(key)) == NULL)
		return (OPT_FAILED);
	*out = e->value;
	return (OPT_OK);
}

const struct option_ops option_ops = {
	get_bool_option,
	get_char_option,
	get_short_option,
	get_int_option,
	get_long_option,
	get_long_long_option,
	get_str_option,
};

const struct option_ops *
opts_init(void)
{
	opt_store.refcount++;
	return (&option_ops);
}

/*
 * opts_fini.  Return 0 if this is the last close and all the options have
 * been used.
 * Return a positive int if this is the last close but there are unused
 * options.
 * Return a negative int if this is not the last close.
 */
int
opts_fini()
{
	if (--opt_store.refcount == 0) {
		report_back_opts();
		return (check_and_free_option_store());
	}
	return (opt_store.refcount * -1);
}

int
load_ilist(struct ilist *ilp, int *vals, int len)
{
	int i;
	if ((ilp->vals = malloc(len * sizeof (uint32_t))) == NULL) {
		MALLOC_ERROR((long)sizeof (uint32_t));
		return (1);
	}
	ilp->weightings = NULL;
	ilp->wlen = len;
	for (i = 0; i < len; i++) {
		ilp->vals[i] = vals[i];
	}
	return (0);
}

/*
 * Set the diskomizer path.  This is the etc directory based on the root of
 * the installation. If we are relocated then we need to try and work out
 * where it has been installed.
 *
 * We expect the path to be $ROOT/bin/$(uname -i)/binary and from that have
 * the path be $ROOT/etc.
 */
const char *
set_diskomizer_path(void)
{
	const char *p;
	char *x;
	char *wx;
	int i;

	if ((p = getenv(DISKOMIZER_PATH)) != NULL) {
		return (p);
	}

	if ((p = getexecname()) == NULL) {
		return (def_diskomizer_path);
	}

	/* Is the path fully qualified? */
	if (p[0] != '/') {
		char *wd;

		wd = getcwd(NULL, 1024);
		if (wd == NULL ||
		    (x = (char *)malloc(MAX(strlen(wd) + strlen(p) + 3, 5))) ==
		    NULL) {
			return (def_diskomizer_path);
		}
		sprintf(x, ":%s/%s", wd, p);
		free(wd);
	} else {
		/*
		 * strlen p must be at least 2, both of which will get
		 * deleted so we add 4 to give enough space for ":/etc"
		 * with a traling NUL.
		 */
		if ((x = (char *)malloc(strlen(p) + 4)) == NULL) {
			return (def_diskomizer_path);
		}
		sprintf(x, ":%s", p);
	}
	/* loose the binary name */
	wx = strrchr(x, '/');
	*wx = NULL;
	/* Now loose 2 directories, "$(uname -i)" and "bin" */
	for (i = 0; i < 2; i++) {
		if ((wx = strrchr(x, '/')) != NULL) {
			*wx = NULL;
		} else {
			break;
		}
	}
	strcat(x, "/etc");
	/* Now reduce the memory used to what is really required */
	if ((wx = realloc(x, strlen(x) + 1)) != NULL) {
		x = wx;
	}
	return (x);
}
