/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

#pragma ident	"@(#)daio_posix.c	1.21	09/05/26 SMI"

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <diskomizer/daio.h>
#include <aio.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <diskomizer/tnf.h>
#include <sys/systeminfo.h>
#include <strings.h>
#include <diskomizer/daio_dev.h>
#include <diskomizer/assert.h>
#include "args.h"
#include "findap.h"

#if defined(__SunOS_5_8) || defined(__SunOS_5_9)

/*
 * Various definitions required to allow the event port code to compile
 * but not run on 5.8 and 5.9.
 *
 * Shamelessly stolen from the 5.10 header files.
 */

#ifndef PORT_SOURCE_AIO
#define	PORT_SOURCE_AIO		1
#endif

#ifndef SIGEV_PORT
#define	SIGEV_PORT		4
#endif

typedef struct port_event {
	int		portev_events;	/* event data is source specific */
	ushort_t	portev_source;	/* event source */
	ushort_t	portev_pad;	/* port internal use */
	uintptr_t	portev_object;	/* source specific object */
	void		*portev_user;	/* user cookie */
} port_event_t;

typedef	struct	port_notify {
	int		portnfy_port;	/* bind request(s) to port */
	void		*portnfy_user;	/* user defined */
} port_notify_t;

int	port_create(void);
int	port_get(int, port_event_t *, struct timespec *);
int	port_getn(int, port_event_t [], uint_t, uint_t *, struct timespec *);

#pragma weak port_create
#pragma weak port_get
#pragma weak port_getn

#else
#include <port.h>
#endif


struct daio_posix {
	int next;
	int error;
	/* return count is only useful for debugging */
	int return_count:4;
	int is_read:1;
	hrtime_t start_time;
	hrtime_t end_time;
	daio_result_t *resultp;
	struct daio_id *id;
	port_notify_t   pn;
	struct aiocb aiocb;
};

/*
 * XOR this value into the sival_int so that this sival_int is zero. This
 * allows diskomizer to spot bug 4520934 as the sival_int will never be zero.
 */
#define	SIGVAL_INTMASK active_list_max

TNF_DECLARE_RECORD(struct daio_posix, daio_posix_str);
TNF_DEFINE_RECORD_4(struct daio_posix, daio_posix_str,
		tnf_uint, aiocb.aio_fildes,
		tnf_uint, aiocb.aio_nbytes,
		tnf_opaque, aiocb.aio_buf,
		tnf_opaque, aiocb.aio_offset)

static struct daio_posix *daio_posix_base;
static struct aiocb const **active_list;
static int active_list_max;
static int active_count = 0;

struct error {
	int sig;
	int exit;
	int entry;
} error;
/*
 * There are three lists of io structures.
 * 1: The free_list containing unused io structures.
 * 2: The siglist, containing ios that have been reaped by the signal
 *	handler. This list is protected by calls to sigprocmask, when
 *	accessed from outside of the signal handler. This is not used when
 *	using ports.
 * 3: The done_list.  This contains ios reaped by the signal handler but
 * 	not yet reaped by the wait routine.  When this list is empty this
 *	filled again by copying the siglist to it, saving the number of times
 *	the signal mask has to be changed. fast as changes to the signal
 *	handler are, this still buys quite a bit or performance when you are
 *	on the edge. This is also not used when using ports.
 */
struct list {
	sigset_t mask;
	int head;
	int tail;
};
static struct list siglist;
static struct list done_list;
static struct list free_list;

static int sigrtmin;
static int port;

typedef enum {
	NEVER = 0,
	USE_PORTS_CHILD, /* create each port in the child processes. */
	USE_PORTS_PARENT /* create one shared port in the master process */
} use_ports_t;

static use_ports_t use_ports_enum;

static void *daio_posix_init(int count);
static void daio_posix_init_checker(const char *checker,
	off64_t max_block_size);
static data_checker_t daio_posix_what_checker(void);
static ssize_t daio_posix_pwrite(int fildes, const uchar_t *bufp, size_t bufs,
		off_t off, struct daio_id *id);
static ssize_t daio_posix_pread(int fildes, uchar_t *bufp, size_t bufs,
		off_t off, struct daio_id *id);
static int daio_posix_awrite(int fildes, uchar_t *bufp, int bufs, off_t offset,
	daio_result_t *resultp, struct daio_id *id);
static int daio_posix_aread(int fildes, uchar_t *bufp, int bufs, off_t offset,
	daio_result_t *resultp, struct daio_id *id);
static daio_status_t daio_posix_status(daio_result_t *resultp);
static hrtime_t daio_posix_start_time(daio_result_t *resultp);
static hrtime_t daio_posix_end_time(daio_result_t *resultp);
static daio_result_t *daio_posix_wait(const struct timeval *timeout);
static daio_result_t *daio_posix_sigwait(const struct timeval *timeout);
static daio_result_t *daio_posix_portwait(const struct timeval *timeout);
static int daio_posix_cancel(daio_result_t *resultp);
static void daio_posix_fini(void *handle);
static int (*data_check)(uchar_t *buf, size_t bufs, struct daio_id *id);
static daio_result_t *(*do_daio_posix_wait)(const struct timeval *timeout);

struct daio_ops DAIO_OPS = {
	daio_posix_init,
	daio_posix_init_checker,
	daio_posix_what_checker,
	findap,
	(daio_open_t)daio_dev_open,
	daio_dev_close,
	unlink,
	stat64,
	fstat64,
	ioctl,
	ftruncate64,
	daio_dev_directio,
	read_disko_vtoc,
	daio_posix_pwrite,
	daio_posix_pread,
	daio_posix_awrite,
	daio_posix_aread,
	daio_posix_status,
	daio_posix_start_time,
	daio_posix_end_time,
	daio_posix_wait,
	daio_posix_cancel,
	daio_posix_fini,
	daio_dev_dd
};

static int
is56()
{
	char buf[257];

	if (sysinfo(SI_RELEASE, buf, sizeof (buf)) == -1) {
		return (1);
	} else if (strcmp(buf, "5.6") == 0) {
		return (1);
	} else {
		return (0);
	}
}

static int
none_active_port(void)
{
	return (active_count == 0 ? 1 : 0);
}

static int
none_active_sig(void)
{
	int i;
	int ret = 1;

	if (done_list.head == -1) {
		sigprocmask(SIG_BLOCK, &siglist.mask, NULL);
		if (siglist.head == -1) {
			for (i = 0; i < active_list_max; i++) {
				if (active_list[i]) {
					ret = 0;
					break;
				}
			}
		} else {
			done_list.head = siglist.head;
			done_list.tail = siglist.tail;
			siglist.head = siglist.tail = -1;
			ret = 0;
		}
		sigprocmask(SIG_UNBLOCK, &siglist.mask, NULL);
	} else {
		ret = 0;
	}
	return (ret);
}
static int
bounds_check(daio_result_t *res)
{
	long x;

	if ((x = (long)res->private_data) >= 0 && x < active_list_max &&
	    daio_posix_base[x].resultp == res) {
		return (x);
	} else {
		return (-1);
	}
}
static void
push_nosig(struct list *l, int entry)
{
	assert(daio_posix_base[entry].next == -1);
	if (l->tail == -1) {
		assert(l->head == -1);
		l->tail = entry;
	} else {
		daio_posix_base[l->head].next = entry;
	}
	daio_posix_base[entry].next = l->head;
	l->head = entry;
}
static void
push(struct list *l, int entry, int inhandler)
{
	if (!inhandler) {
		sigprocmask(SIG_BLOCK, &l->mask, NULL);
		push_nosig(l, entry);
		sigprocmask(SIG_UNBLOCK, &l->mask, NULL);
	} else {
		push_nosig(l, entry);
	}
}
static int
pop_nosig(struct list *l)
{
	int x;
	x = l->tail;
	if (x == l->head) {
		l->tail = l->head = -1;
	} else {
		l->tail = daio_posix_base[x].next;
	}
	daio_posix_base[x].next = -1;
	return (x);
}
static int
pop(struct list *l)
{
	int x;

	sigprocmask(SIG_BLOCK, &l->mask, NULL);
	x = pop_nosig(l);
	sigprocmask(SIG_UNBLOCK, &l->mask, NULL);
	return (x);
}
#define	ADD_TO_SIG(X, Y) \
		assert(daio_posix_base[X].error != EINPROGRESS); \
		assert(siglist.head == -1 || \
		daio_posix_base[siglist.head].aiocb.aio_buf != NULL); \
		push(&siglist, X, Y)
/*
 * The free list is only manipulated in the read, write and wait routines
 * so does not need protecting from signals
 */
#define	ADD_TO_FREE(X) push_nosig(&free_list, X)

static int
get_free(void)
{
	return (pop_nosig(&free_list));
}
static int
pop_from_done(void)
{
	int i;


	i = done_list.head;
	if (i != -1) {
		assert(daio_posix_base[i].error != EINPROGRESS);
		assert(daio_posix_base[done_list.head].aiocb.aio_buf !=
		    NULL);
		i = pop_nosig(&done_list);
		daio_posix_base[i].next = -1;
	} else {
		sigprocmask(SIG_BLOCK, &siglist.mask, NULL);
		done_list.head = siglist.head;
		done_list.tail = siglist.tail;
		siglist.head = siglist.tail = -1;
		sigprocmask(SIG_UNBLOCK, &siglist.mask, NULL);
		if (-1 != done_list.head) {
			i = pop_nosig(&done_list);
		}
	}
	return (i);
}
static void
add_to_active(int entry)
{
	assert(daio_posix_base[entry].next == -1);
	active_list[entry] = &daio_posix_base[entry].aiocb;
	daio_posix_base[entry].error = EINPROGRESS;
	daio_posix_base[entry].start_time = gethrtime();
	active_count++;
}
static int
check_error(int i)
{
	if (daio_posix_base[i].error == EINPROGRESS) {
		daio_posix_base[i].error =
		    aio_error(&daio_posix_base[i].aiocb);
	}
	return (daio_posix_base[i].error);
}
#define	is_active(entry) (active_list[entry] != NULL)

static void
remove_from_active(int entry)
{
	assert(active_list[entry] != NULL);
	daio_posix_base[entry].end_time = gethrtime();
	active_list[entry] = NULL;
	active_count--;
}

static use_ports_t
use_ports(void)
{
	return (use_ports_enum);
}

/*ARGSUSED*/
static void
daio_posix_handler(int signo, siginfo_t *siginfo, void *context)
{
	int entry = siginfo->si_value.sival_int ^ SIGVAL_INTMASK;
	int x;

	/* Have to work around 4120502 */
	if (siginfo->si_code != SI_ASYNCIO) {
		error.sig++;
		return;
	}

	if (entry < 0 || entry >= SIGVAL_INTMASK ||
	    active_list[entry] == NULL) {
		error.exit++;
		error.entry = entry;
		return;
	}
	x = check_error(entry);
	if (x != EINPROGRESS && x != ECANCELED) {
		daio_posix_base[entry].return_count++;
		remove_from_active(entry);
		ADD_TO_SIG(entry, 1);
	}
}

static void
daio_posix_opts(void)
{
	const struct option_ops *ops;
	char x;

#if defined(__SunOS_5_8) || defined(__SunOS_5_9)
	/*
	 * Check the weak symbols are actually available.
	 */
	if (port_create == NULL || port_get == NULL || port_getn == NULL) {
		return;
	}
#endif

	ops = opts_init();
	if (ops->opt_bool("DAIO_POSIX_USE_PORTS", &x) == OPT_OK && x == 1) {
		use_ports_enum = USE_PORTS_CHILD;
		if (ops->opt_bool("DAIO_POSIX_SHARE_PORT", &x) == OPT_OK &&
		    x == 1) {
			use_ports_enum = USE_PORTS_PARENT;
			port = port_create();
		}
	} else {
		use_ports_enum = USE_PORTS_CHILD;
	}
	opts_fini();
}

static void *
daio_posix_init(int count)
{
	int i;
	struct sigaction action;
	sigrtmin = SIGRTMIN;

	free_list.head = free_list.tail = -1;
	done_list.head = done_list.tail = -1;
	siglist.head = siglist.tail = -1;

	if (!use_ports()) {

		sigemptyset(&free_list.mask);
		sigemptyset(&done_list.mask);
		sigemptyset(&siglist.mask);
		sigaddset(&siglist.mask, sigrtmin);

		action.sa_flags = SA_SIGINFO;
		do_daio_posix_wait = daio_posix_sigwait;
		action.sa_sigaction = daio_posix_handler;
		sigemptyset(&action.sa_mask);
		(void) sigaction(sigrtmin, &action, NULL);
	} else {
		do_daio_posix_wait = daio_posix_portwait;
		if (use_ports() != 2) {
			port = port_create();
		}
		if (port == -1) {
			return (NULL);
		}
	}


	if ((daio_posix_base =
		calloc(count, sizeof (struct daio_posix))) == NULL) {
		return (NULL);
	}
	if ((active_list = calloc(count, sizeof (struct aiocb *))) == NULL) {
		free(daio_posix_base);
		return (NULL);
	}
	for (i = 0; i < count; i++) {
		daio_posix_base[i].next = -1;
		ADD_TO_FREE(i);
	}

	active_list_max = count;

	return (daio_posix_base);
}

/*ARGSUSED*/
static void
daio_posix_init_checker(const char *checker, off64_t max_block_size)
{
	daio_posix_opts();

	data_check = choose_data_checker(checker);
}
static data_checker_t
daio_posix_what_checker(void)
{
	return (data_check);
}
/*ARGSUSED*/
static ssize_t
daio_posix_pwrite(int fildes, const uchar_t *bufp, size_t bufs, off_t off,
		struct daio_id *id)
{
	return (pwrite(fildes, bufp, bufs, off));
}
static ssize_t
daio_posix_pread(int fildes, uchar_t *bufp, size_t bufs, off_t off,
		struct daio_id *id)
{
	ssize_t ret;

	ret = pread(fildes, bufp, bufs, off);

	if (id != NULL && ret == bufs) {
		if (data_check(bufp, ret, id) < 0)
			ret = DAIO_CORRUPT;
	}
	return (ret);
}
static void
set_notification(int this_one)
{
	if (use_ports()) {

		daio_posix_base[this_one].pn.portnfy_port = port;
		daio_posix_base[this_one].pn.portnfy_user = (void *)this_one;
		daio_posix_base[this_one].aiocb.aio_sigevent.sigev_notify =
		    SIGEV_PORT;
		daio_posix_base[this_one
		    ].aiocb.aio_sigevent.sigev_value.sival_ptr =
		    &daio_posix_base[this_one].pn;

	} else {
		daio_posix_base[this_one].aiocb.aio_sigevent.sigev_notify =
		    SIGEV_SIGNAL;
		daio_posix_base[this_one].aiocb.aio_sigevent.sigev_signo =
		    sigrtmin;
		daio_posix_base[this_one
		    ].aiocb.aio_sigevent.sigev_value.sival_int =
		    this_one ^ SIGVAL_INTMASK;
	}
}

static int
daio_posix_aread(int fildes, uchar_t *bufp, int bufs, off_t offset,
	daio_result_t *resultp, struct daio_id *id)
{
	int this_one;
	int status;

	if (free_list.head == -1) {
		return (-1);
	}
	this_one = get_free();

	daio_posix_base[this_one].id = id;
	daio_posix_base[this_one].return_count = 0;
	daio_posix_base[this_one].is_read = 1;
	daio_posix_base[this_one].aiocb.aio_buf = bufp;
	daio_posix_base[this_one].aiocb.aio_fildes = fildes;
	daio_posix_base[this_one].aiocb.aio_nbytes = bufs;
	daio_posix_base[this_one].aiocb.aio_offset = offset;

	set_notification(this_one);
	assert(bufp);
	add_to_active(this_one);
	TNF_PROBE_2(aio_read, "aio_read", "sunw%cte%diskomizer%posix_aio read",
	    tnf_uint, this_one, this_one,
	    daio_posix_str, &daio_posix_base[this_one],
	    &daio_posix_base[this_one]);
	if ((status = aio_read(&daio_posix_base[this_one].aiocb)) == -1) {
		remove_from_active(this_one);
		ADD_TO_FREE(this_one);
	} else {
		resultp->private_data = (void *) ((long)this_one);
		daio_posix_base[this_one].resultp = resultp;
	}
	return (status);
}
static int
daio_posix_awrite(int fildes, uchar_t *bufp, int bufs, off_t offset,
	daio_result_t *resultp, struct daio_id *id)
{
	int this_one;
	int status;

	if (free_list.head == -1) {
		return (-1);
	}
	this_one = get_free();

	assert(bufp);
	daio_posix_base[this_one].id = id;
	daio_posix_base[this_one].return_count = 0;
	daio_posix_base[this_one].is_read = 0;
	daio_posix_base[this_one].aiocb.aio_buf = bufp;
	daio_posix_base[this_one].aiocb.aio_fildes = fildes;
	daio_posix_base[this_one].aiocb.aio_nbytes = bufs;
	daio_posix_base[this_one].aiocb.aio_offset = offset;

	set_notification(this_one);

	add_to_active(this_one);
	TNF_PROBE_2(aio_write, "aio_write",
	    "sunw%cte%diskomizer%posix_aio write",
	    tnf_uint, this_one, this_one,
	    daio_posix_str, &daio_posix_base[this_one],
	    &daio_posix_base[this_one]);
	if ((status = aio_write(&daio_posix_base[this_one].aiocb)) == -1) {
		remove_from_active(this_one);
		ADD_TO_FREE(this_one);
	} else {
		resultp->private_data = (void *)((long)this_one);
		daio_posix_base[this_one].resultp = resultp;
	}
	return (status);
}
static daio_status_t
daio_posix_status(daio_result_t *resultp)
{
	int x;

	if ((x = bounds_check(resultp)) == -1) {
		return (DAIO_COMPLETE);
	}

	if (is_active(x)) {
		return (DAIO_INPROGRESS);
	}
	return (DAIO_COMPLETE);
}
static void
check_exit(void)
{
	/*
	 * No need to fiddle with the signal mask unless there is an error.
	 * If there are errors to report they will get caught next time round.
	 */
	if (error.exit || error.sig) {
		sigset_t mask;
		sigemptyset(&mask);
		sigaddset(&mask, sigrtmin);
		sigprocmask(SIG_BLOCK, &mask, NULL);
		if (error.sig)  {
			plog(LOG_WARNING,
			    "%d SIGRTMIN signal%s received not from aio,"
			    " ignored.\n",
			    error.sig, error.sig > 1 ? "s" : "");
			error.sig = 0;
		}
		if (error.exit) {
			plog(LOG_ERR, "invalid aio signal for entry %d\n",
			    error.entry);
			sigprocmask(SIG_UNBLOCK, &mask, NULL);
			exit(1);
		}
		sigprocmask(SIG_UNBLOCK, &mask, NULL);
	}
}

daio_result_t *
daio_posix_wait(const struct timeval *timeout)
{
	return (do_daio_posix_wait(timeout));
}

daio_result_t *
daio_posix_sigwait(const struct timeval *timeout)
{
	int this_one;
	daio_result_t *resultp;

	if (daio_posix_base == NULL) {
		errno = EINVAL;
		return ((daio_result_t *)-1);
	}

	check_exit();

	if ((this_one = pop_from_done()) == -1) {
		if (timeout && !none_active_sig()) {
			(void) poll(NULL, 0, (timeout->tv_sec * 1000) +
			    timeout->tv_usec/1000);
			this_one = pop_from_done();
		}
		if (this_one == -1) {
			if (none_active_sig()) {
				errno = EINVAL;
				return ((daio_result_t *)-1);
			} else {
				return ((daio_result_t *)NULL);
			}
		}
	}
	resultp = daio_posix_base[this_one].resultp;
	daio_posix_base[this_one].resultp = NULL;
	resultp->result.daio_errno = daio_posix_base[this_one].error;
	resultp->result.daio_return =
	    aio_return(&daio_posix_base[this_one].aiocb);
	resultp->result.daio_start_time = daio_posix_base[this_one].start_time;
	resultp->result.daio_end_time = daio_posix_base[this_one].end_time;
	if (daio_posix_base[this_one].is_read &&
	    resultp->result.daio_return ==
	    daio_posix_base[this_one].aiocb.aio_nbytes &&
	    data_check((uchar_t *)daio_posix_base[this_one].aiocb.aio_buf,
	    resultp->result.daio_return, daio_posix_base[this_one].id) < 0) {
		resultp->result.daio_return = DAIO_CORRUPT;
	}
	assert(daio_posix_base[this_one].error != EINPROGRESS);

	TNF_PROBE_4(aio_wait, "aio_wait",
	    "sunw%cte%diskomizer%posix_aio pwait",
	    tnf_uint, this_one, this_one,
	    tnf_uint, daio_return, resultp->result.daio_return,
	    tnf_uint, daio_errno, resultp->result.daio_errno,
	    daio_posix_str, &daio_posix_base[this_one],
	    &daio_posix_base[this_one]);

	ADD_TO_FREE(this_one);

	return (resultp);
}

daio_result_t *
daio_posix_portwait(const struct timeval *timeout)
{
	port_event_t ev;
	timespec_t *tsp;
	timespec_t ts;
	daio_result_t *resultp;
	int this_one;

	if (none_active_port()) {
		errno = EINVAL;
		return ((daio_result_t *)-1);
	}

	if (timeout != NULL) {
		ts.tv_sec = timeout->tv_sec;
		ts.tv_nsec = timeout->tv_usec * 1000;
		tsp = &ts;
	} else {
		tsp = NULL;
	}

	if (port_get(port, &ev, tsp) == -1) {
		return (NULL);
	}
	if (ev.portev_source != PORT_SOURCE_AIO) {
		return (NULL); /* can't happen */
	}

	this_one = (int)ev.portev_user;

	if ((check_error(this_one) == EINPROGRESS)) {
		return (NULL); /* can't happen */
	}
	daio_posix_base[this_one].return_count++;
	remove_from_active(this_one);

	resultp = daio_posix_base[this_one].resultp;
	daio_posix_base[this_one].resultp = NULL;
	resultp->result.daio_errno = daio_posix_base[this_one].error;
	resultp->result.daio_return =
	    aio_return(&daio_posix_base[this_one].aiocb);
	resultp->result.daio_start_time = daio_posix_base[this_one].start_time;
	resultp->result.daio_end_time = daio_posix_base[this_one].end_time;
	if (daio_posix_base[this_one].is_read &&
	    resultp->result.daio_return ==
	    daio_posix_base[this_one].aiocb.aio_nbytes &&
	    data_check((uchar_t *)daio_posix_base[this_one].aiocb.aio_buf,
	    resultp->result.daio_return, daio_posix_base[this_one].id) < 0) {
		resultp->result.daio_return = DAIO_CORRUPT;
	}
	assert(daio_posix_base[this_one].error != EINPROGRESS);

	TNF_PROBE_4(aio_wait, "aio_wait",
	    "sunw%cte%diskomizer%posix_aio pwait",
	    tnf_uint, this_one, this_one,
	    tnf_uint, daio_return, resultp->result.daio_return,
	    tnf_uint, daio_errno, resultp->result.daio_errno,
	    daio_posix_str, &daio_posix_base[this_one],
	    &daio_posix_base[this_one]);

	ADD_TO_FREE(this_one);

	return (resultp);
}

static int
daio_posix_cancel(daio_result_t *resultp)
{
	int io;
	int x;

	if ((io = bounds_check(resultp)) == -1 || !is_active(io)) {
		return (-1);
	}
	x = aio_cancel(daio_posix_base[io].aiocb.aio_fildes,
	    &daio_posix_base[io].aiocb);
	if (x == AIO_CANCELED) {
		assert(check_error(io) == ECANCELED);
	}
	return (x == AIO_CANCELED ? 0 : -1);
}

/*ARGSUSED*/
static void
daio_posix_fini(void *handle)
{
	if (use_ports()) {
		close(port);
	}
	free(active_list);
	free(daio_posix_base);
}

static hrtime_t
daio_posix_start_time(daio_result_t *resultp)
{
	int x;

	if ((x = bounds_check(resultp)) == -1) {
		return (DAIO_NOT_STARTED);
	}
	return (daio_posix_base[x].start_time);
}
static hrtime_t
daio_posix_end_time(daio_result_t *resultp)
{
	int x;

	if ((x = bounds_check(resultp)) == -1) {
		return (DAIO_NOT_STARTED);
	}
	return (daio_posix_base[x].end_time);
}
