/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

#pragma ident	"@(#)daio_sunos.c	1.16	09/05/26 SMI"

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <diskomizer/daio.h>
#include <diskomizer/checker.h>
#include "findap.h"
#include <sys/asynch.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <diskomizer/daio_dev.h>
#include <diskomizer/assert.h>

static void *daio_sunos_init(int count);
static void daio_sunos_init_checker(const char *checker,
	off64_t max_block_size);
static data_checker_t daio_sunos_what_checker(void);
static ssize_t daio_sunos_pwrite(int fildes, const uchar_t *bufp, size_t bufs,
		off_t off, struct daio_id *id);
static ssize_t daio_sunos_pread(int fildes, uchar_t *bufp, size_t bufs,
		off_t off, struct daio_id *id);
static int daio_sunos_awrite(int fildes, uchar_t *bufp, int  bufs, off_t
	offset, daio_result_t *resultp, struct daio_id *id);
static int daio_sunos_aread(int fildes, uchar_t *bufp, int  bufs, off_t
	offset, daio_result_t *resultp, struct daio_id *id);
static daio_status_t daio_sunos_status(daio_result_t *resultp);
static hrtime_t daio_sunos_start_time(daio_result_t *resultp);
static hrtime_t daio_sunos_end_time(daio_result_t *resultp);
static daio_result_t *daio_sunos_wait(const struct timeval *tv);
static int daio_sunos_cancel(daio_result_t *resultp);

static void daio_sunos_fini(void *handle);

static int (*data_check)(uchar_t *buf, size_t bufs, struct daio_id *id);

struct daio_sunos_str {
	daio_result_t res;
	daio_result_t *resp;
	int next;
	struct daio_id *id;
	uchar_t *bufp;
};
struct daio_ops DAIO_OPS = {
	daio_sunos_init,
	daio_sunos_init_checker,
	daio_sunos_what_checker,
	findap,
	(daio_open_t)daio_dev_open,
	daio_dev_close,
	unlink,
	stat64,
	fstat64,
	ioctl,
	ftruncate64,
	daio_dev_directio,
	read_disko_vtoc,
	daio_sunos_pwrite,
	daio_sunos_pread,
	daio_sunos_awrite,
	daio_sunos_aread,
	daio_sunos_status,
	daio_sunos_start_time,
	daio_sunos_end_time,
	daio_sunos_wait,
	daio_sunos_cancel,
	daio_sunos_fini,
	daio_dev_dd
};

struct daio_sunos_str *daio_sunos_base;
int daio_size = -1;
int free_list = -1;

static int
push(int list, int entry)
{
	assert(entry < daio_size);
	assert(free_list < daio_size);
	daio_sunos_base[entry].next = list;
	return (entry);
}
#define	pop(list) (list = daio_sunos_base[list].next)
#define	ADD_TO_FREE(X) free_list = push(free_list, X)
static int
get_free(void)
{
	int entry = free_list;
	assert(free_list < daio_size);
	pop(free_list);
	assert(free_list < daio_size);
	daio_sunos_base[entry].next = -1;
	return (entry);
}
static int
bounds_check(daio_result_t *res)
{
	int x = (int)((long)res->private_data);

	if (x >= 0 && x < daio_size && daio_sunos_base[x].resp == res) {
		return (x);
	} else {
		return (-1);
	}
}

static void *
daio_sunos_init(int count)
{
	int i;
	daio_sunos_base = calloc(count, sizeof (struct daio_sunos_str));
	if (daio_sunos_base == NULL) {
		return (NULL);
	}
	daio_size = count;
	for (i = 0; i < count; i++) {
		daio_sunos_base[i].next = -1;
		ADD_TO_FREE(i);
	}

	return (daio_sunos_base);
}

/*ARGSUSED*/
static void
daio_sunos_init_checker(const char *checker, off64_t max_block_size)
{
	data_check = choose_data_checker(checker);
}

static data_checker_t
daio_sunos_what_checker(void)
{
	return (data_check);
}

static int
daio_sunos_awrite(int fildes, uchar_t *bufp, int  bufs, off_t
	offset, daio_result_t *resultp, struct daio_id *id)
{
	int i;
	int result;
	assert(free_list < daio_size);
	i = get_free();

	daio_sunos_base[i].id = id;
	daio_sunos_base[i].res = *resultp;
	daio_sunos_base[i].resp = resultp;
	daio_sunos_base[i].bufp = NULL;
	resultp->private_data = (void *)((long)i);
	DAIO_SET_RETURN(daio_sunos_base[i].res, AIO_INPROGRESS);
	DAIO_SET_START_TIME(daio_sunos_base[i].res, gethrtime());

	assert(free_list < daio_size);

	result = aiowrite64(fildes, (char *)bufp, bufs, offset, SEEK_SET,
	    (aio_result_t *)&daio_sunos_base[i].res);

	if (result != 0) {
		ADD_TO_FREE(i);
	}
	return (result);
}
static int
daio_sunos_aread(int fildes, uchar_t *bufp, int  bufs, off_t
	offset, daio_result_t *resultp, struct daio_id *id)
{
	int i;
	int result;
	i = get_free();
	assert(free_list < daio_size);

	daio_sunos_base[i].id = id;
	daio_sunos_base[i].res = *resultp;
	daio_sunos_base[i].resp = resultp;
	daio_sunos_base[i].bufp = bufp;
	resultp->private_data = (void *)((long)i);
	DAIO_SET_RETURN(daio_sunos_base[i].res, AIO_INPROGRESS);
	DAIO_SET_START_TIME(daio_sunos_base[i].res, gethrtime());

	assert(free_list < daio_size);

	result = aioread64(fildes, (char *)bufp, bufs, offset, SEEK_SET,
	    (aio_result_t *)&daio_sunos_base[i].res);

	if (result != 0) {
		ADD_TO_FREE(i);
	}
	return (result);
}
static daio_status_t
daio_sunos_status(daio_result_t *resultp)
{
	int x;
	assert(free_list < daio_size || daio_size == -1);
	if ((x = bounds_check(resultp)) == -1) {
		return (DAIO_COMPLETE);
	} else {
		return (DAIO_RETURN(daio_sunos_base[x].res) == AIO_INPROGRESS ?
		    DAIO_INPROGRESS : DAIO_COMPLETE);
	}
}
static daio_result_t *
daio_sunos_wait(const struct timeval *timeout)
{
	struct daio_sunos_str *resp;
	daio_result_t *resultp;
	assert(free_list < daio_size || daio_size == -1);
	resp = (struct daio_sunos_str *)aiowait((struct timeval *)timeout);
	if (resp != NULL && resp != (struct daio_sunos_str *)-1) {
		resultp = resp->resp;
		*resultp = resp->res;
		resp->resp = NULL;
		DAIO_SET_END_TIME(*resultp, gethrtime());

		if (resp->bufp != NULL &&
		    DAIO_RETURN(*resultp) == resp->id->bufs &&
		    data_check(resp->bufp, DAIO_RETURN(*resultp),
		    resp->id) < 0) {
			DAIO_SET_RETURN(*resultp, DAIO_CORRUPT);
		}
		assert(&daio_sunos_base[resp - daio_sunos_base] == resp);
		ADD_TO_FREE(resp - daio_sunos_base);
	} else {
		resultp = (daio_result_t *)resp;
	}
	return (resultp);
}
static int
daio_sunos_cancel(daio_result_t *resultp)
{
	struct daio_sunos_str *resp;
	int ret;
	int i;

	assert(free_list < daio_size);

	if (daio_sunos_status(resultp) == DAIO_INPROGRESS) {
		errno = EACCES;
		return (-1);
	}
	i = (int)((long)resultp->private_data);

	resp = &daio_sunos_base[i];
	ret = aiocancel((aio_result_t *)&resp->res);

	if (ret == 0) {
		*resultp = resp->res;
		ADD_TO_FREE(i);
	}
	return (ret);
}
/*ARGSUSED*/
static void
daio_sunos_fini(void *handle)
{
	if (daio_sunos_base && daio_sunos_base == handle) {
		free(daio_sunos_base);
		daio_sunos_base = NULL;
	}
}

static hrtime_t
daio_sunos_start_time(daio_result_t *resultp)
{
	int i = (int)((long)resultp->private_data);
	assert(free_list < daio_size);
	return (DAIO_GET_START_TIME(daio_sunos_base[i].res));
}

static hrtime_t
daio_sunos_end_time(daio_result_t *resultp)
{
	int i = (int)((long)resultp->private_data);
	assert(free_list < daio_size);
	return (DAIO_GET_END_TIME(daio_sunos_base[i].res));
}

/*ARGSUSED*/
static ssize_t
daio_sunos_pwrite(int fildes, const uchar_t *bufp, size_t bufs, off_t off,
		struct daio_id *id)
{
	return (pwrite(fildes, bufp, bufs, off));
}

static ssize_t
daio_sunos_pread(int fildes, uchar_t *bufp, size_t bufs, off_t off,
		struct daio_id *id)
{
	ssize_t ret;

	ret = pread(fildes, bufp, bufs, off);

	if (id != NULL && ret == bufs) {
		if (data_check(bufp, ret, id) < 0)
			ret = DAIO_CORRUPT;
	}
	return (ret);
}
