/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

#pragma ident	"@(#)daio_uscsi.c	1.13	09/05/26 SMI"

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <sys/scsi/impl/uscsi.h>
#include <sys/scsi/generic/commands.h>
#include <sys/scsi/impl/commands.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>

#include <diskomizer/daio_dev.h>
#include <diskomizer/daio.h>
#include <diskomizer/log.h>
#include "findap.h"
#include "daio_async.h"
#include "args.h"

#ifdef BIG_ENDIAN
#define	CONV2INT(A) (A)
#else
#define	SWABB(X) (((0xff & (X)) << 8) | (0xff00 & (X)) >> 8)
#define	SWABS(X) ((SWABB(0xffff & (X)) << 16) |\
		SWABB((0xffff0000 & (X)) >> 16))
#define	CONV2INT(A) SWABS(A)
#endif

#define	MAX_GROUP2OFF 0xffffffffLL

/*
 * Group 4 Commands, Direct Access Devices
 *
 * The ifndef is required as on later releases these are or should be
 * in the header files.
 */
#ifndef SCMD_READ_G4
#define	SCMD_READ_G4		0x88
#endif

#ifndef SCMD_WRITE_G4
#define	SCMD_WRITE_G4		0x8a
#endif

#ifndef SCMD_WRITE_VERIFY_G4
#define	SCMD_WRITE_VERIFY_G4	0x8e
#endif

#ifndef FORMG4LONGADDR
#define	FORMG4LONGADDR(cdb, addr) \
	(cdb)->g4_addr3 = (addr) >> 56; \
	(cdb)->g4_addr2 = ((addr) >> 48) & 0xFF; \
	(cdb)->g4_addr1 = ((addr) >> 40) & 0xFF; \
	(cdb)->g4_addr0 = ((addr) >> 32) & 0xFF; \
	(cdb)->g4_addtl_cdb_data3 = ((addr) >> 24) & 0xFF; \
	(cdb)->g4_addtl_cdb_data2 = ((addr) >> 16) & 0xFF; \
	(cdb)->g4_addtl_cdb_data1 = ((addr) >> 8) & 0xFF; \
	(cdb)->g4_addtl_cdb_data0 = (addr) & 0xFF
#endif


union lun_n_tag {
	uchar_t data;
	struct {
#if defined(_BIT_FIELDS_LTOH)
		uchar_t reladdr:1;
		uchar_t bytchk:1;
		uchar_t res:1;
		uchar_t fua:1;
		uchar_t dpo:1;
		uchar_t lun:3;
#elif defined(_BIT_FIELDS_HTOL)
		uchar_t lun:3;
		uchar_t dpo:1;
		uchar_t fua:1;
		uchar_t res:1;
		uchar_t bytchk:1;
		uchar_t reladdr:1;
#else
#error One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif
	} un;
};

struct timeout_vals {
	uint16_t (*set_timeout)(struct timeout_vals *v);
	int16_t min;
	int16_t diff;
};

static union lun_n_tag wlun_n_tag;
static union lun_n_tag rlun_n_tag;

static int uscsi_open(const char *path, int oflag, mode_t mode);
static int uscsi_close(int fd);
static void *uscsi_read_disko_vtoc(int fd);
static ssize_t uscsi_pread(int fildes, void *buf, size_t nbyte, off_t offset);
static ssize_t uscsi_pwrite(int fildes, void *buf, size_t nbyte, off_t offset);
static int daio_uscsi_write(int fildes, uchar_t *bufp, int bufs, off_t
		offset, daio_result_t *resultp, struct daio_id *id);
static int daio_uscsi_read(int fildes, uchar_t *bufp, int bufs, off_t
		offset, daio_result_t *resultp, struct daio_id *id);
static int nodev(int, int);

struct daio_ops DAIO_OPS = {
	daio_async_init,
	daio_async_init_checker,
	daio_async_get_checker,
	findap,
	(daio_open_t)uscsi_open,
	uscsi_close,
	unlink,
	stat64,
	fstat64,
	ioctl,
	ftruncate64,
	nodev, /* drect io does not make sense for USCSI */
	uscsi_read_disko_vtoc,
	daio_async_pwrite,
	daio_async_pread,
	daio_uscsi_write,
	daio_uscsi_read,
	daio_async_status,
	daio_async_start_time,
	daio_async_end_time,
	daio_async_wait,
	daio_async_cancel,
	daio_async_fini,
	daio_dev_dd
};

struct uscsi_str {
	void *handle;
	uint_t v_sectorsz;
	struct disko_partition part;
};

static struct uscsi_str *uscsi_state;
static int maxfd = -1;

static struct timeout_vals uscsi_read_timeout;
static struct timeout_vals uscsi_write_timeout;
static int8_t write_cmd1;
static int8_t write_cmd4;
static int8_t allow_bad_vtoc;
static int8_t ignore_partitions = 0;
ssize_t (*pwrite_func)(int fildes, void *bufp, size_t bufs,
	off_t off) = uscsi_pwrite;
ssize_t (*pread_func)(int fildes, void *bufp, size_t bufs,
	off_t off) = uscsi_pread;

static int
nodev(int x, int b)
{
	return (0);
}

static uint16_t
set_timeout_diff(struct timeout_vals *v)
{
	return (v->min + (lrand48() % v->diff));
}

static uint16_t
set_timeout_no_diff(struct timeout_vals *v)
{
	return (v->min);
}

static void
io_timeouts(const struct option_ops *ops, char *base_str,
	struct timeout_vals *v)
{
	char *str;

	if (ops->opt_short(base_str, &v->min) == OPT_OK) {
		/* for comatability */
		plog(LOG_NOTICE, "%s %d second%s\n",
		    str, v->min, v->min == 1 ? "" : "s");
		v->diff = 0;
	} else if ((str = malloc(strlen(base_str) + 5)) != NULL) {
		int16_t max;

		sprintf(str, "%s.min", base_str);
		if (ops->opt_short(str, &v->min) != OPT_OK) {
			v->min = 60;
		} else {
			plog(LOG_NOTICE, "%s %d second%s\n",
			    str, v->min, v->min == 1 ? "" : "s");
		}
		sprintf(str, "%s.max", base_str);
		if (ops->opt_short(str, &max) != OPT_OK) {
			max = 60;
		} else {
			plog(LOG_NOTICE, "%s %d second%s\n",
			    str, max, max == 1 ? "" : "s");
		}
		if (max < v->min) {
			plog(LOG_WARNING,
			    "%s max (%d) less then min (%d)"
			    " setting min to max\n", base_str, max, v->min);
			v->min = max;
			v->diff = 0;
		} else {
			v->diff = max - v->min;
		}
		free(str);
	} else {
		v->min = 60;
		v->diff = 0;
	}
	if (v->diff == 0) {
		v->set_timeout = set_timeout_no_diff;
	} else {
		v->set_timeout = set_timeout_diff;
	}
}

static void
uscsi_init(void)
{
	const struct option_ops *ops;
	char x;

	ops = opts_init();

	plog(LOG_NOTICE, "DAIO_USCSI version %s\n", VERSION);

	if (ops->opt_bool("DAIO_USCSI_IGNORE_PARTITIONS", &x) ==
	    OPT_OK && x == 1) {
		plog(LOG_NOTICE, "USCSI IGNORE_PARTITIONS on\n");
		ignore_partitions = 1;
	}

	if (ops->opt_bool("DAIO_USCSI_USE_PWRITE", &x) == OPT_OK && x == 1) {
		if (ignore_partitions) {
			plog(LOG_ERR, "DAIO_USCSI_USE_PWRITE and "
			    "DAIO_USCSI_IGNORE_PARTITIONS are mutually "
			    "exclusive\n");
			exit(1);
		}
		pwrite_func =
		    (ssize_t (*)(int, void *, size_t, off_t)) pwrite64;
	} else {

		if (ops->opt_bool("DAIO_USCSI_WRITE_VERIFY", &x) == OPT_OK &&
		    x == 1) {
			write_cmd1 = SCMD_WRITE_VERIFY;
			write_cmd4 = SCMD_WRITE_VERIFY_G4;
			plog(LOG_NOTICE, "USCSI WRITE_VERIFY on\n");
			if (ops->opt_bool("DAIO_USCSI_WRITE_VERIFY_BYTE_CHECK",
			    &x) == OPT_OK && x == 1) {
				plog(LOG_NOTICE,
				    "USCSI WRITE VERIFY_BYTE_CHECK on\n");
				wlun_n_tag.un.bytchk = 1;
			}
		} else {
			write_cmd1 = SCMD_WRITE_G1;
			write_cmd4 = SCMD_WRITE_G4;
		}

		if (ops->opt_bool("DAIO_USCSI_WRITE_FORCE_UNIT_ACCESS", &x) ==
		    OPT_OK && x == 1) {
			plog(LOG_NOTICE,
			    "USCSI WRITE FORCE_UNIT_ACCESS on\n");
			wlun_n_tag.un.fua = 1;
		}

		if (ops->opt_bool("DAIO_USCSI_WRITE_DISABLE_PAGE_OUT", &x) ==
		    OPT_OK && x == 1) {
			plog(LOG_NOTICE,
			    "USCSI WRITE DISABLE_PAGE_OUT on\n");
			wlun_n_tag.un.dpo = 1;
		}

		io_timeouts(ops, "DAIO_USCSI_WRITE_TIMEOUT",
		    &uscsi_write_timeout);
	}

	if (ops->opt_bool("DAIO_USCSI_USE_PREAD", &x) == OPT_OK && x == 1) {
		if (ignore_partitions) {
			plog(LOG_ERR, "DAIO_USCSI_USE_PREAD and "
			    "DAIO_USCSI_IGNORE_PARTITIONS are mutually "
			    "exclusive\n");
			exit(1);
		}
		pread_func = pread64;
	} else {
		if (ops->opt_bool("DAIO_USCSI_READ_FORCE_UNIT_ACCESS", &x) ==
		    OPT_OK && x == 1) {
			plog(LOG_NOTICE, "USCSI READ_UNIT_ACCESS on\n");
			rlun_n_tag.un.fua = 1;
		}

		if (ops->opt_bool("DAIO_USCSI_READ_DISABLE_PAGE_OUT", &x) ==
		    OPT_OK && x == 1) {
			plog(LOG_NOTICE,
			    "USCSI READ DISABLE_PAGE_OUT on\n");
			rlun_n_tag.un.dpo = 1;
		}

		io_timeouts(ops, "DAIO_USCSI_READ_TIMEOUT",
		    &uscsi_read_timeout);
	}

	if (ops->opt_bool("DAIO_USCSI_ALLOW_BAD_VTOC", &x) ==
	    OPT_OK && x == 1) {
		plog(LOG_NOTICE, "USCSI ALLOW_BAD_VTOC on\n");
		allow_bad_vtoc = 1;
	}

	opts_fini();
}

static int
uscsi_read_capacity(int fildes, uint32_t *bs, uint64_t *mba)
{
	struct uscsi_cmd ucmd;
	union scsi_cdb cdb;
	struct buf {
		uint32_t mba;
		uint32_t bs;
	} buf;

	memset(&ucmd, 0, sizeof (struct uscsi_cmd));
	memset(&cdb, 0, sizeof (cdb));
	memset(&buf, 0, sizeof (buf));

	cdb.scc_cmd = SCMD_READ_CAPACITY;
	ucmd.uscsi_cdblen = CDB_GROUP1;
	ucmd.uscsi_flags = USCSI_READ;
	ucmd.uscsi_cdb = (char *)&cdb;
	ucmd.uscsi_bufaddr = (void *)&buf;
	ucmd.uscsi_buflen = sizeof (buf);
	ucmd.uscsi_resid = sizeof (buf);

	if (ioctl(fildes, USCSICMD, &ucmd) == -1) {
		return (-1);
	}

	if (ucmd.uscsi_status != 0) {
		errno = EIO;
		return (-1);
	}
	*bs = CONV2INT(buf.bs);
	*mba = CONV2INT(buf.mba);

	return (ucmd.uscsi_resid);
}

static int
uscsi_open(const char *path, int oflag, mode_t mode)
{
	int fd;
	if (uscsi_read_timeout.min == 0 && uscsi_read_timeout.diff == 0) {
		uscsi_init();
	}

	if ((fd = daio_dev_open(path, oflag, mode)) == -1) {
		return (-1);
	}
	if (fd > maxfd) {
		struct uscsi_str *tmp;

		tmp = realloc(uscsi_state, (fd+1) * sizeof (struct uscsi_str));
		if (tmp == NULL) {
			daio_dev_close(fd);
			errno = ENOMEM;
			return (-1);
		}
		memset(&tmp[maxfd+1], 0,
		    (fd - maxfd) * sizeof (struct uscsi_str));
		uscsi_state = tmp;
		maxfd = fd;
	}
	uscsi_state[fd].handle = uscsi_read_disko_vtoc(fd);
	plog(LOG_NOTICE, "USCSI %s\n", path);
	plog(LOG_NOTICE, "USCSI sector size %d disk size %ld\n",
	    uscsi_state[fd].v_sectorsz, uscsi_state[fd].part.p_size);
	return (fd);
}

static int
uscsi_close(int fd)
{
	if (uscsi_state[fd].handle) {
		disko_vtoc_free(uscsi_state[fd].handle);
	}
	memset(&uscsi_state[fd], 0, sizeof (uscsi_state[fd]));
	return (daio_dev_close(fd));
}

static struct disko_vtoc *
uscsi_read_capacity_vtoc(int fd)
{
	struct disko_vtoc *dv;

	if ((dv = alloc_disko_vtoc(1)) == NULL) {
		return (NULL);
	}
	memset(&dv->v_volume[0], 0, sizeof (dv->v_volume));
	memset(&dv->v_asciilabel[0], 0, sizeof (dv->v_asciilabel));

	uscsi_state[fd].part.p_start = 0;
	dv->v_sectorsz = uscsi_state[fd].v_sectorsz;
	dv->v_nparts = 1;
	dv->this_part = 0;
	dv->v_part[0] = uscsi_state[fd].part;
	plog(LOG_NOTICE, "USCSI read_capacity: file %s",
	    daio_dev_path(fd));
	plog(LOG_NOTICE, "USCSI read_capacity: "
	    "Number of sectors %lld sector size %d\n",
	    uscsi_state[fd].part.p_size, uscsi_state[fd].v_sectorsz);
	return (dv);
}

static void *
uscsi_read_disko_vtoc(int fd)
{
	void *handle;

	if ((handle = uscsi_state[fd].handle) != NULL) {
		/*
		 * We are no longer in cotrol of the memory associated with
		 * the handle. It will be freed by the caller so drop our
		 * copy of it. If we were to need the vtoc from the disk
		 * again we can get it from the disk.
		 *
		 * In practice we don't.
		 */
		uscsi_state[fd].handle = NULL;
		return (handle);
	}
	if (ignore_partitions) {
		return (uscsi_read_capacity_vtoc(fd));
	}

	if ((handle = read_disko_vtoc(fd)) == NULL) {
		struct disko_vtoc *dv;
		if (!allow_bad_vtoc) {
			return (NULL);
		}
		return (uscsi_read_capacity_vtoc(fd));
	}
	uscsi_state[fd].v_sectorsz = disko_vtoc_sectorsz(handle);
	uscsi_state[fd].part = *disko_vtoc_this_partition(handle);
	return (handle);
}

static ssize_t
uscsi_prw(int fildes, void *buf, size_t nbyte, off_t offset, int cmd)
{
	struct uscsi_cmd ucmd;
	uint64_t lba;
	uint32_t len;
	union scsi_cdb cdb;

	if (nbyte % uscsi_state[fildes].v_sectorsz != 0 ||
	    offset % uscsi_state[fildes].v_sectorsz != 0) {
		plog(LOG_ERR,
		    "USCSI io at 0x%llx length 0x%x to file %s not aligned "
		    "with 0x%x\n",
		    offset, nbyte, daio_dev_path(fildes),
		    uscsi_state[fildes].v_sectorsz);
		errno = EINVAL;
		return (-1);
	}
	lba = offset/uscsi_state[fildes].v_sectorsz;
	len = nbyte/uscsi_state[fildes].v_sectorsz;

	if (!ignore_partitions) {
		if (lba + len > uscsi_state[fildes].part.p_size) {
			errno = ENXIO;
			return (-1);
		}
		lba += uscsi_state[fildes].part.p_start;
	}
	if (len > 256) {
		plog(LOG_ERR,
		    "USCSI io at %lld length %d (%d) to file %s to big "
		    "for bs %d\n",
		    offset, nbyte, len, daio_dev_path(fildes),
		    uscsi_state[fildes].v_sectorsz);
		errno = EINVAL;
		return (-1);
	}

	memset(&ucmd, 0, sizeof (struct uscsi_cmd));
	memset(&cdb, 0, sizeof (cdb));

	if (cmd == USCSI_READ) {
		if (lba > MAX_GROUP2OFF) {
			cdb.scc_cmd = SCMD_READ_G4;
		} else {
			cdb.scc_cmd = SCMD_READ_G1;
		}
		cdb.scc_lun = rlun_n_tag.data;
		ucmd.uscsi_timeout =
		    uscsi_read_timeout.set_timeout(&uscsi_read_timeout);
	} else {
		if (lba > MAX_GROUP2OFF) {
			cdb.scc_cmd = write_cmd4;
		} else {
			cdb.scc_cmd = write_cmd1;
		}
		cdb.scc_lun = wlun_n_tag.data;
		ucmd.uscsi_timeout =
		    uscsi_write_timeout.set_timeout(&uscsi_write_timeout);
	}

	if (lba > MAX_GROUP2OFF) {
		FORMG4LONGADDR(&cdb, lba);
		FORMG4COUNT(&cdb, len);
		ucmd.uscsi_cdblen = CDB_GROUP4;
	} else {
		uint32_t lba32 = (uint32_t)lba;
		FORMG1ADDR(&cdb, lba32);
		FORMG1COUNT(&cdb, len);
		ucmd.uscsi_cdblen = CDB_GROUP1;
	}
	ucmd.uscsi_flags = cmd;
	ucmd.uscsi_cdb = (char *)&cdb;
	ucmd.uscsi_bufaddr = buf;
	ucmd.uscsi_buflen = nbyte;
	ucmd.uscsi_resid = nbyte;

	if (ioctl(fildes, USCSICMD, &ucmd) == -1) {
		return (-1);
	}

	if (ucmd.uscsi_status != 0) {
		plog(LOG_ERR,
		    "USCSI io at %x length %x to file %s status %d\n",
		    offset, nbyte, daio_dev_path(fildes),
		    ucmd.uscsi_status);
		errno = EIO;
		return (-1);
	}

	if (ucmd.uscsi_resid) {
		plog(LOG_ERR, "nbyte %d ucmd.uscsi_resid %d\n",
		    nbyte, ucmd.uscsi_resid);
	}
	errno = 0;
	return (nbyte - ucmd.uscsi_resid);
}

static ssize_t
uscsi_pread(int fildes, void *buf, size_t nbyte, off_t offset)
{
	return (uscsi_prw(fildes, buf, nbyte, offset, USCSI_READ));
}

static ssize_t
uscsi_pwrite(int fildes, void *buf, size_t nbyte, off_t offset)
{
	return (uscsi_prw(fildes, buf, nbyte, offset, USCSI_WRITE));
}

static int
daio_uscsi_read(int fildes, uchar_t *bufp, int bufs, off_t offset,
	daio_result_t *resultp, struct daio_id *id)
{
	return (daio_async_rw(fildes, bufp, bufs, offset, resultp, id,
	    pread_func, IS_READ));
}
static int
daio_uscsi_write(int fildes, uchar_t *bufp, int bufs, off_t offset,
	daio_result_t *resultp, struct daio_id *id)
{
	return (daio_async_rw(fildes, bufp, bufs, offset, resultp, id,
	    pwrite_func, IS_WRITE));
}
