/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)daio.h	1.16	09/05/26 SMI"

#ifndef _DAIO_H
#define	_DAIO_H

#ifdef	__cplusplus
extern "C" {
#endif

/*
 * Check that the compilation configuration is correct.
 */
#if ! defined(_LP64)
#if !defined(_LARGEFILE_SOURCE)
#error "_LARGEFILE_SOURCE must be set"
#endif
#if _FILE_OFFSET_BITS != 64
#error "_FILE_OFFSET_BITS must be 64"
#endif
#endif

#define	DAIO_OPS v8_daio_ops
#define	DAIO_OPS_STRING "v8_daio_ops"

/*
 * diskomizer aio framework.
 *
 * This is a framework to allow different aio schemes to be used in
 * diskomizer with minimum run time impact.
 */

#include <diskomizer/tnf.h>
#include <diskomizer/vtoc.h>
#include <sys/types.h>
#include <tnf/probe.h>
#include <sys/stat.h>

#define	DAIO_CORRUPT -2

enum  daio_status {
	DAIO_ERROR = -2,	/* A DAIO Internal error */
	DAIO_INPROGRESS = -1,
	DAIO_COMPLETE = 0,
	DAIO_COMPLETE_OK = 1,
	DAIO_COMPLETE_ERROR = 2
};
typedef enum daio_status daio_status_t;
struct daio_aio_result {
	ssize_t daio_return;
	int daio_errno;
	/* Both these times are to be set by the daio implementation.  */
	hrtime_t daio_start_time; /* Time when the io was given to the OS */
	hrtime_t daio_end_time; /* Time when the io was completed by the OS */
};
typedef struct daio_aio_result daio_aio_result_t;
TNF_DECLARE_RECORD(daio_aio_result_t, daio_tnf_aio_result);

typedef uint64_t check_t;

/*
 * daio_id structure. Passed to both daio->read and daioi->write.  For
 * writes it describes the buffer being written. The buf_id is guaranteed
 * to be unique to this buffer. (NOTE the buffer address is not
 * guaranteed to be unique).
 *
 * When passed to daio->read it contains the information about the last
 * write to the block being written. This allows the daio library to check
 * the data that was read is good. The id is the unique identifier for the
 * buffer, the buf is the buffer that was written (the header in that buffer
 * is not the header that was written. The header is the one supplied in
 * hdr). The same is true for the footer.
 */
struct daio_id {
	int32_t buf_id;
	int32_t bufs;
	int32_t hdr_len;
	int32_t footer_len;
	check_t chksum;
	uchar_t *hdr;
	uchar_t *footer;
	uchar_t *buf;
};

#include <diskomizer/checker.h>

/*
 * This should be considered opaque to the caller.
 */
union daio_result_t {
	struct daio_aio_result result;
	void *private_data; /* to bu used by the individual implementations */
};
typedef union daio_result_t daio_result_t;

TNF_DECLARE_RECORD(daio_result_t, daio_tnf_result);

typedef int (*daio_open_t)(const char *path, int oflag, /* mode_t mode */...);

struct daio_ops {
	/*
	 * Give the init routine the number of simultaneous aio operations
	 * that you will do. Memory will be allocated by the init routine
	 * so that there is no run time cost of memory allocation and
	 * deallocation. If you need more aios later then you can do
	 * a new init to get some more.
	 *
	 * block_size is the maximum block size that will be used.
	 */
	void * (*init)(int count);
	/*
	 * Master init routine.  This one gets run before diskomizer forks.
	 *
	 * Checker is a string that describe the kind of integrity
	 * checking to do.
	 */
	void (*init_master)(const char *checker, off64_t block_size);
	/*
	 * What checker the checker in use by daio.
	 */
	data_checker_t (*what_checker)(void);
	/*
	 * Findap finds all the alternate paths for this daio instance.
	 * The format of the arguments is entirely defined by the daio
	 * implementation in use.
	 */
	struct paths *(*findap)(const char *path, const char *devices);
	/*
	 * open, close and stat. The path argument is defined by the daio
	 * implentation in use, the return value similarly so and need
	 * not refer to an open file descriptor. It's use with any routine
	 * that is not part of the daio is undefined. The oflag and mode
	 * take the same behaviour as in the SunOS open() routine.
	 */
	int (*open)(const char *path, int oflag, /* mode_t mode */...);
	int (*close)(int);
	int (*unlink)(const char *path);
	int (*stat)(const char *path, struct stat64 *buf);
	/*
	 * All the rest of the routines take a fildes as returned by the
	 * open routine above. Any other fildes results are undefined.
	 */
	int (*fstat)(int fildes, struct stat64 *buf);
	/*
	 * Ioctl entry point to implement the dkio(7I) calls.
	 */
	int (*ioctl)(int fildes, int request, ...);
	int (*ftruncate)(int fildes, off64_t length);
	int (*directio)(int fd, int advice);
	/*
	 * read_vtoc. Reads an opaque vtoc which can be queried with the
	 * routines in vtoc.so.
	 */
	void *(*read_vtoc)(int fildes);
	/*
	 * pwrite and pread routines.  The id in both cases can be NULL
	 * but apart from that these are the same as the awrite and aread
	 * routines in their relationship with the SunOS counter parts.
	 */
	ssize_t (*pwrite)(int fildes, const uchar_t *bufp, size_t bufs,
		off_t off, struct daio_id *id);
	ssize_t (*pread)(int fildes, uchar_t *bufp, size_t bufs, off_t off,
		struct daio_id *id);
	/*
	 * Write and read options. take the same arguments as the SunOS
	 * aioread and aiowrite routines, except there is no whence flag
	 * all ios are SEEK_SET and there is an extra id option.
	 *
	 * In both cases the id is the id of the buffer that is being
	 * written to the device.
	 *
	 */
	int (*awrite)(int fildes, uchar_t *bufp, int  bufs, off_t
		offset, daio_result_t *resultp,
		struct daio_id *id);
	/*
	 * read also checks the data, returning -2 via daio->wait
	 * if the error has been
	 * corrupted. It only has to fill in bufb if there was data
	 * corruption so that the caller can print further diganostics in
	 * the case of success the contents of bufp is undefined. It may have
	 * been written to or it may not.
	 */
	int (*aread)(int fildes, uchar_t *bufp, int  bufs, off_t
		offset, daio_result_t *resultp,
		struct daio_id *id);
	daio_status_t (*status)(daio_result_t *resultp);
	/*
	 * Start time can only be called with a resultp that has been
	 * passed to read or write but not yet returned by wait.
	 *
	 * It is allowed to return DAIO_NOT_STARTED
	 */
#define	DAIO_NOT_STARTED -17
	hrtime_t (*start_time)(daio_result_t *resultp);
	hrtime_t (*end_time)(daio_result_t *resultp);
	daio_result_t *(*wait)(const struct timeval *timeout);
	int (*cancel)(daio_result_t *resultp);
	void (*fini)(void * handle); /* pass back the handle got from init */
	/*
	 * Routine to print a command to read the data that is corrupt.
	 * Typically the command is dd(1) but that need not be the case.
	 */
	void (*plog_dd)(int pri, int fildes, size_t len, size_t bs,
		off_t offset);
};
/*
 * These macros can ONLY be used after the aio has been returned from wait
 * or within the daio routines themselves.
 * At any other time the results are undefined.
 */
#define	DAIO_RETURN(A) ((A).result.daio_return)
#define	DAIO_ERROR(A) ((A).result.daio_errno)
#define	DAIO_GET_START_TIME(A) ((A).result.daio_start_time)
#define	DAIO_GET_END_TIME(A) ((A).result.daio_end_time)
#define	DAIO_GET_TIME_TAKEN(A)	(DAIO_GET_END_TIME(A) - \
	DAIO_GET_START_TIME(A))
/*
 * Macros used to fake a DAIO
 */
#define	DAIO_SET_RETURN(A, B) ((A).result.daio_return = B)
#define	DAIO_SET_ERROR(A, B) ((A).result.daio_errno = B)
#define	DAIO_SET_START_TIME(A, B) ((A).result.daio_start_time = B)
#define	DAIO_SET_END_TIME(A, B) ((A).result.daio_end_time = B)

extern struct daio_ops *daio_choose_ops(const char *name);


#ifdef	__cplusplus
}
#endif

#endif /* _DAIO_H */
