/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)recent.c	1.5	09/05/26 SMI"

/*
 * Recent writes.
 *
 *	The idea here is to keep a per process, per device list of offests
 *	that have been recently written. This is so that when an event
 *	occurs diskomizer can quickly check that the last few IO's to
 *	go complete were good.
 *
 *	The intention is that while the blocks are still being held as
 *	recently written, they are locked and therefore not able to be
 *	written again.  The down side of this is that they are not elegible
 *	to be read either.
 *
 *	This can be turned on by setting the size of the recent write log
 *	to zero.
 */

#include "diskomizer64mpism.h"

static ullong_t retoff(struct recent_blocks *blks, ullong_t off);
static ullong_t minus1(struct recent_blocks *blks);
static void *zero(void);
ullong_t (*push_recent)(struct recent_blocks *blks, ullong_t off) = retoff;
ullong_t (*pop_recent)(struct recent_blocks *blks) = minus1;
int (*snapshot_recent)(struct recent_blocks *blks) =
	(int (*)(struct recent_blocks *blks))zero;

static ullong_t
retoff(struct recent_blocks *blks, ullong_t x)
{
	return (x);
}
static ullong_t
minus1(struct recent_blocks *blks)
{
	return ((ullong_t)-1);
}
static void *
zero(void)
{
	return (NULL);
}
/*
 * Push the offset onto the list of recent writes.  If the list is full
 * return the offset of the oldest io so that it can be freed
 */
ullong_t
do_push_recent(struct recent_blocks *blks, ullong_t off)
{
	ullong_t retoff;

	retoff = blks->recent_blocks[blks->first];
	blks->recent_blocks[blks->last] = off;
	blks->last = (blks->last + 1) % blks->length;
	if (blks->last == blks->first) {
		blks->first = (blks->first + 1) % blks->length;
		return (retoff);
	}
	return (-1);
}
/*
 * snapshot the recently written blocks, leaving the log empty.
 */
int
do_snapshot_recent(struct recent_blocks *blks)
{
	int x = blks->frozen_block_count;
	int i;

	blks->frozen_block_count = 0;

	for (i = blks->first; i != blks->last; i = (i + 1) % blks->length) {
		blks->recent_blocks[blks->length + blks->frozen_block_count++] =
		    blks->recent_blocks[i];
		blks->recent_blocks[i] = -1;
	}
	blks->first = i;
	return (x);
}
/*
 * pop an entry from the  snapshot
 */
ullong_t
do_pop_recent(struct recent_blocks *blks)
{
	if (blks->frozen_block_count == 0)
		return ((ullong_t)-1);
	return (blks->recent_blocks[blks->length + --blks->frozen_block_count]);
}
struct recent_blocks *
init_recent(short count)
{
	struct recent_blocks *blks;

	if (count <= 0)
		return (NULL);
	blks = malloc(sizeof (struct recent_blocks) +
	    ((count - 1) * 2 * sizeof (ullong_t)));
	if (blks == NULL)
		return (NULL);
	blks->length = count;
	blks->first = 0;
	blks->last = 0;
	blks->recent_blocks[0] = -1;
	push_recent = do_push_recent;
	snapshot_recent = do_snapshot_recent;
	pop_recent = do_pop_recent;
	return (blks);
}
