/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 *	Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 *	Use is subject to license terms.
 */

#pragma ident	"@(#)usage_tracking.c	1.8	09/05/26 SMI"

#include <sys/types.h>
#include <sys/systeminfo.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <time.h>
#include <limits.h>
#include <signal.h>

#include <usage_tracking/usage_tracking.h>

#ifdef ASSERT_USAGE_TRACKING
#include <diskomizer/assert.h>
#else
#define	assert(A)
#endif

#define	SYSINFO_BUF_SIZE 25

struct key_list {
	char *key;
	struct key_list *next;
};

typedef struct usage_track {
	struct key_list *list;
	char *sendmail;
	char *to;
	char *from;
	char *tool;
	char *version;
	char *dom;
	char *str;
	int len;
	int indent;
	int errors;
} usage_track_t;

static const char xml_version[] = "<?xml version=\"1.0\"?>\n";

#define	NOT_NULL_FREE(X) if (X != NULL) { free(X); X = NULL; }

static struct key_list *
push_key(usage_track_t *usage_trackp, const char *key)
{
	struct key_list *x;

	if ((x = malloc(sizeof (struct key_list))) == NULL) {
		return (NULL);
	}
	if ((x->key = strdup(key)) == NULL) {
		free(x);
		return (NULL);
	}
	x->next = usage_trackp->list;
	usage_trackp->list = x;
	return (x);
}

static char *
pop_key(usage_track_t *usage_trackp)
{
	char *x;
	struct key_list *l;

	if ((l = usage_trackp->list) == NULL) {
		return (NULL);
	}
	x = l->key;
	usage_trackp->list = l->next;
	free(l);

	return (x);
}

static void
free_usage_track(usage_track_t *usage_trackp)
{
	char *key;

	NOT_NULL_FREE(usage_trackp->to);
	NOT_NULL_FREE(usage_trackp->from);
	NOT_NULL_FREE(usage_trackp->sendmail);
	NOT_NULL_FREE(usage_trackp->dom);
	NOT_NULL_FREE(usage_trackp->str);
	NOT_NULL_FREE(usage_trackp->tool);
	NOT_NULL_FREE(usage_trackp->version);

	while ((key = pop_key(usage_trackp)) != NULL) {
		free(key);
	}
	free(usage_trackp);
}


static void
close_open_keys(usage_track_t *usage_trackp)
{
	while (usage_trackp->list != NULL) {
		usage_tracking_close_key((void *)usage_trackp);
	}
}

static void
indent(usage_track_t *usage_trackp)
{
	int i;

	for (i = 0; i < usage_trackp->indent; i++) {
		usage_trackp->str[usage_trackp->len++] = '\t';
	}
}

void
usage_tracking_store_key_value(void *handle,
		const char *key, const char *value)
{
	/*
	 * No need to push and pop the key as we are just doing the whole
	 * thing in one hit.
	 */
	int this_len;
	char *tmp;
	usage_track_t *usage_trackp = handle;

	if (usage_trackp == NULL) {
		return;
	}

	this_len = (3 * usage_trackp->indent) + (2 * strlen(key)) +
	    strlen(value) + 9;

	if ((tmp = realloc(usage_trackp->str,
	    usage_trackp->len + this_len + 1)) == NULL) {
		usage_trackp->errors++;
	} else {
		int x =  usage_trackp->len;

		usage_trackp->str = tmp;
		indent(usage_trackp);
		usage_trackp->len +=
		    sprintf(&usage_trackp->str[usage_trackp->len],
		    "<%s>\n", key);

		indent(usage_trackp);

		usage_trackp->len +=
		    sprintf(&usage_trackp->str[usage_trackp->len],
		    "\t%s\n", value);
		indent(usage_trackp);

		usage_trackp->len +=
		    sprintf(&usage_trackp->str[usage_trackp->len],
		    "</%s>\n", key);

		assert(x + this_len ==  usage_trackp->len);
	}
}

void
usage_tracking_store_key_value_int(void *handle,
		const char *key, const long long value)
{
	char buf[64];
	sprintf(buf, "%lld", value);
	usage_tracking_store_key_value(handle, key, buf);
}

int
usage_tracking_open_key(void *handle,
	const char *key, const char *version, const char *id)
{
	int this_len;
	char *tmp;
	usage_track_t *usage_trackp = handle;

	if (usage_trackp == NULL) {
		return (0);
	}

	this_len = strlen(key) + /* "<>\n" */  3 + usage_trackp->indent;

	if (version != NULL) {
		this_len += strlen(version) + /* " version=\"\"" */ 11;
	}

	if (id != NULL) {
		this_len += strlen(id) + /* " id=\"\"" */ 6;
	}

	if ((tmp = realloc(usage_trackp->str,
	    usage_trackp->len + this_len + 1)) == NULL) {
		usage_trackp->errors++;
		return (0);
	} else {
		int x =  usage_trackp->len;
		int i;

		if (push_key(usage_trackp, key) == NULL) {
			usage_trackp->errors++;
			return (0);
		}

		usage_trackp->str = tmp;

		indent(usage_trackp);

		i = sprintf(&usage_trackp->str[usage_trackp->len], "<%s", key);
		usage_trackp->len += i;

		if (version != NULL) {
			i = sprintf(&usage_trackp->str[usage_trackp->len],
			    " version=\"%s\"", version);
			usage_trackp->len += i;
		}

		if (id != NULL) {
			i = sprintf(&usage_trackp->str[usage_trackp->len],
			    " id=\"%s\"", id);
			usage_trackp->len += i;
		}

		i = sprintf(&usage_trackp->str[usage_trackp->len], ">\n");
		usage_trackp->len += i;

		usage_trackp->indent++;

		assert(x + this_len ==  usage_trackp->len);
	}
	return (1);
}

int
usage_tracking_close_key(void *handle)
{
	int this_len;
	int ret;
	char *key;
	char *tmp;
	usage_track_t *usage_trackp = handle;

	if (usage_trackp == NULL) {
		return (0);
	}

	if ((key = pop_key(usage_trackp)) == NULL ||
	    --usage_trackp->indent < 0) {
		usage_trackp->indent = 0;
		usage_trackp->errors++;
		usage_tracking_store_key_value(handle, "usage_track_error",
		    "Unbalanced close");
		return (0);
	}
	this_len = strlen(key) + /* "</>\n" */ 4 + usage_trackp->indent;

	if ((tmp = realloc(usage_trackp->str,
	    usage_trackp->len + this_len + 1)) == NULL) {
		usage_trackp->errors++;
		ret = 0;
	} else {
		int x =  usage_trackp->len;

		usage_trackp->str = tmp;

		indent(usage_trackp);

		sprintf(&usage_trackp->str[usage_trackp->len],
		    "</%s>\n", key);
		usage_trackp->len += strlen(key) + 4;
		assert(x + this_len ==  usage_trackp->len);
		ret = 1;
	}
	free(key);
	return (ret);
}

void
usage_tracking_store_value(void *handle,
		const char *value)
{
	int this_len;
	char *tmp;
	usage_track_t *usage_trackp = handle;

	if (usage_trackp == NULL) {
		return;
	}

	this_len = strlen(value) + /* "\n" */  1 + usage_trackp->indent;

	if ((tmp = realloc(usage_trackp->str,
	    usage_trackp->len + this_len + 1)) == NULL) {
		usage_trackp->errors++;
		return;
	} else {
		int x =  usage_trackp->len;
		usage_trackp->str = tmp;

		indent(handle);

		usage_trackp->len +=
		    sprintf(&usage_trackp->str[usage_trackp->len],
		    "%s\n", value);
		assert(x + this_len ==  usage_trackp->len);
	}
}

static int
usage_tracking_ok(const char *dom, const char *host)
{
	char *buf;
	char *tmp;
	int dom_len = strlen(dom);
	long len = 0;
	long ret;
	char x[1];

	if (dom_len == 0 || (ret = sysinfo(SI_SRPC_DOMAIN, x, 1)) <= dom_len) {
		return (0);
	}
	if ((buf = malloc(ret)) == NULL) {
		return (0);
	}
	len = ret;

	if ((ret = sysinfo(SI_SRPC_DOMAIN, buf, len)) > len) {
		free(buf);
		return (0);
	}

	ret--; /* remove the space for the NULL from the length */

	tmp = &buf[ret - dom_len];

	if (strcasecmp(tmp, dom) != 0) {
		free(buf);
		return (0);
	}
	free(buf);
	return (1);
}

static char *
now_str(char *str)
{
	time_t now = time(NULL);

	cftime(str, "%D %T %z", &now);
	return (str);
}

static void
usage_tracking_start_time(usage_track_t *usage_trackp)
{
	if (usage_tracking_open_key(usage_trackp, "START_TIME", NULL, NULL)) {
		time_t now = time(NULL);
		char buf[128];
		int len;

		ctime_r(&now, buf, sizeof (buf));

		len = strlen(buf) - 1;

		if (buf[len] = '\n') {
			buf[len] = NULL;
		}

		usage_tracking_store_key_value(usage_trackp,
		    "LOCAL_TIME", buf);
		usage_tracking_store_key_value_int(usage_trackp,
		    "UNIX_TIME", now);

		usage_tracking_close_key((void *)usage_trackp);
	}
}

static void
usage_tracking_sysinfo(usage_track_t *usage_trackp,
	int command, const char *str)
{
	char buf[SYSINFO_BUF_SIZE];
	char *t;
	int x;

	if ((x = sysinfo(command, buf, sizeof (buf))) > sizeof (buf)) {
		if ((t = malloc(x)) != NULL) {
			(void) sysinfo(command, t, x);
		} else {
			t = buf;
		}
	} else {
		t = buf;
	}

	usage_tracking_store_key_value(usage_trackp, str, t);

	if (t != buf) {
		free(t);
	}

}


static void
usage_tracking_system_info(usage_track_t *usage_trackp)
{
	if (usage_tracking_open_key(usage_trackp, "SYSTEM_INFO",
	    NULL, NULL)) {

		usage_tracking_sysinfo(usage_trackp, SI_SYSNAME, "OSNAME");

		usage_tracking_sysinfo(usage_trackp, SI_HOSTNAME, "NODENAME");

		usage_tracking_sysinfo(usage_trackp, SI_SRPC_DOMAIN,
		    "DOMAINNAME");

		usage_tracking_sysinfo(usage_trackp, SI_RELEASE, "OSVERSION");

		usage_tracking_sysinfo(usage_trackp, SI_MACHINE, "MACHINE");

		usage_tracking_sysinfo(usage_trackp, SI_PLATFORM, "PLATFORM");

		usage_tracking_sysinfo(usage_trackp, SI_HW_SERIAL, "SERIAL");

		usage_tracking_sysinfo(usage_trackp, SI_HW_PROVIDER,
		    "HWPROVIDER");

		usage_tracking_sysinfo(usage_trackp, SI_ISALIST,
		    "ISALIST");

		usage_tracking_close_key((void *)usage_trackp);
	}
}

static void
usage_tracking_tool_info(usage_track_t *usage_trackp)
{
	if (usage_tracking_open_key(usage_trackp, "TOOL_INFO", NULL, NULL)) {
		usage_tracking_store_key_value(usage_trackp, "NAME",
		    usage_trackp->tool);
		usage_tracking_store_key_value(usage_trackp, "VERSION",
		    usage_trackp->version);
		usage_tracking_store_key_value_int(usage_trackp,
		    "POINTER_BITS", sizeof (void *) * CHAR_BIT);

		usage_tracking_close_key((void *)usage_trackp);
	}
}


static void
usage_tracking_general(usage_track_t *usage_trackp)
{
	if (usage_tracking_open_key(usage_trackp, "GENERAL", NULL, NULL)) {
		usage_tracking_start_time(usage_trackp);
		usage_tracking_system_info(usage_trackp);
		usage_tracking_tool_info(usage_trackp);

		usage_tracking_close_key((void *)usage_trackp);
	}
}

void *
open_usage_tracking(const char *dom,
	const char *sendmail,
	const char *to,
	const char *from,
	const char *tool,
	const char *version)
{
	usage_track_t *usage_trackp;

	if ((usage_trackp = malloc(sizeof (usage_track_t))) == NULL) {
		return (NULL);
	}

	memset(usage_trackp, 0, sizeof (usage_track_t));

	if ((usage_trackp->dom = strdup(dom)) == NULL ||
	    (usage_trackp->sendmail = strdup(sendmail)) == NULL ||
	    (usage_trackp->from = strdup(from)) == NULL ||
	    (usage_trackp->to = strdup(to)) == NULL ||
	    (usage_trackp->tool = strdup(tool)) == NULL ||
	    (usage_trackp->version = strdup(version)) == NULL) {
		free_usage_track(usage_trackp);
		return (NULL);
	}

	if ((usage_trackp->str = strdup(xml_version)) != NULL) {
		usage_trackp->len = sizeof (xml_version) - 1;
	}

	if (usage_tracking_open_key(usage_trackp,
	    USAGE_TRACKING, USAGE_TRACKING_VERSION, tool)) {

		usage_tracking_general(usage_trackp);

		usage_tracking_open_key(
		    usage_trackp, "TOOLDEFINED", NULL, NULL);
	}

	return (usage_trackp);
}

void
close_usage_tracking(void *handle)
{
	usage_track_t *usage_trackp = handle;

	if (usage_trackp == NULL) {
		return;
	}

	free_usage_track(usage_trackp);
}


void
send_usage_tracking(void *handle)
{
	FILE *fd;
	char *command;
	char now[256];
	sigset_t nset; /* new set */
	sigset_t oset; /* old set */
	int sigprocmask_status;
	usage_track_t *usage_trackp = handle;

	if (usage_trackp == NULL) {
		return;
	}

	now_str(now);

	if (!usage_tracking_ok(usage_trackp->dom, NULL)) {
		return;
	}

	close_open_keys(usage_trackp);

	if ((command = malloc(strlen(usage_trackp->sendmail) +
	    strlen(usage_trackp->to) + 2)) == NULL) {
		return;
	}

	sprintf(command, "%s %s", usage_trackp->sendmail, usage_trackp->to);

	if (access(usage_trackp->sendmail, X_OK) != 0) {
		free(command);
		return;
	}

	fd = popen(command, "w");

	free(command);

	if (fd == NULL) {
		return;
	}

	sigaddset(&nset, SIGPIPE);
	sigprocmask_status = sigprocmask(SIG_BLOCK, &nset, &oset);

	fprintf(fd, "From: %s\nTo: %s\nDate: %s\nSubject: %s\n%s\n.\n",
	    usage_trackp->from,
	    usage_trackp->to, now,
	    usage_trackp->tool, usage_trackp->str);

	pclose(fd);
	if (sigprocmask_status == 0) {
		(void) sigprocmask(SIG_SETMASK, &nset, &oset);
	}
}

void
save_usage_tracking(void *handle, const char *file)
{
	int fd;
	int i;
	int written;
	usage_track_t *usage_trackp = handle;

	if (usage_trackp == NULL) {
		return;
	}

	if (file == NULL ||
	    (fd = open(file, O_WRONLY|O_CREAT|O_TRUNC, 0666)) == -1) {
		return;
	}

	close_open_keys(usage_trackp);

	for (written = 0; written < usage_trackp->len; written += i) {
		i = write(fd, &usage_trackp->str[written],
		    usage_trackp->len - written);
		if (i == -1) {
			break;
		}
	}
	close(fd);
}
